from datetime import datetime

from PyQt5.QtWidgets import QDialog
from PyQt5.QtWidgets import QTreeView
from PyQt5.QtWidgets import QHeaderView
from PyQt5.QtWidgets import QWidget
from PyQt5.QtWidgets import QFormLayout

from PyQt5.QtGui import QStandardItemModel
from PyQt5.QtGui import QBrush
from PyQt5.QtGui import QColor

from PyQt5.QtCore import Qt
from PyQt5.QtCore import QCoreApplication

from galacteek.ipfs.wrappers import ipfsOp
from galacteek.ipfs.cidhelpers import IPFSPath
from galacteek.core.modelhelpers import *

from . import ui_keys, ui_addkeydialog

from .widgets import GalacteekTab
from .helpers import *

from .i18n import iUnknown
from .i18n import iP2PKey


def iKeyName():
    return QCoreApplication.translate('KeysForm', 'Name')


def iKeyResolve():
    return QCoreApplication.translate('KeysForm', 'Resolves to')


class AddKeyDialog(QDialog):
    def __init__(self, app, parent=None):
        super().__init__(parent)

        self.keysView = parent

        self.app = app
        self.ui = ui_addkeydialog.Ui_AddKeyDialog()
        self.ui.setupUi(self)
        self.ui.addKeyLayout.setRowWrapPolicy(QFormLayout.DontWrapRows)
        self.ui.addKeyLayout.setFieldGrowthPolicy(
            QFormLayout.ExpandingFieldsGrow)

    def accept(self):
        keyName = self.ui.keyName.text()
        keySizeText = self.ui.keySize.currentText()

        if len(keyName) == 0:
            return messageBox('Key name is empty')

        self.app.task(self.createKey, keyName, int(keySizeText))

    @ipfsOp
    async def createKey(self, ipfsop, keyName, keySize):
        await ipfsop.client.key.gen(keyName,
                                    type='rsa', size=keySize)
        self.done(1)
        self.keysView.updateKeysList()

    def reject(self):
        self.done(0)


class KeysView(QTreeView):
    def mousePressEvent(self, event):
        if event.button() == Qt.RightButton:
            pass
        QTreeView.mousePressEvent(self, event)


class KeyItem(UneditableItem):
    pass


class KeyResolvedItem(UneditableItem):
    def __init__(self, text):
        super(KeyResolvedItem, self).__init__(text)

        self.resolvedLast = None
        self.resolvesTo = None


class KeysTab(GalacteekTab):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)

        self.resolveTimeout = 60 * 5
        self.keysW = QWidget()
        self.addToLayout(self.keysW)
        self.ui = ui_keys.Ui_KeysForm()
        self.ui.setupUi(self.keysW)

        self.ui.addKeyButton.clicked.connect(self.onAddKeyClicked)
        self.ui.deleteKeyButton.clicked.connect(self.onDelKeyClicked)

        self.model = QStandardItemModel(parent=self)

        self.ui.treeKeys = KeysView()
        self.ui.treeKeys.doubleClicked.connect(self.onItemDoubleClicked)
        self.ui.treeKeys.setModel(self.model)

        self.ui.verticalLayout.addWidget(self.ui.treeKeys)

        self.setupModel()
        self.app.task(self.listKeys)

    def setupModel(self):
        self.model.clear()
        self.model.setColumnCount(3)
        self.model.setHorizontalHeaderLabels([
            iKeyName(), iP2PKey(), iKeyResolve()])
        self.ui.treeKeys.header().setSectionResizeMode(
            0, QHeaderView.ResizeToContents)
        self.ui.treeKeys.header().setSectionResizeMode(
            1, QHeaderView.ResizeToContents)

    def onDelKeyClicked(self):
        idx = self.ui.treeKeys.currentIndex()
        if not idx.isValid():
            return messageBox('Invalid key')

        idxName = self.model.index(idx.row(), 0, idx.parent())
        keyName = self.model.data(idxName)

        if not keyName:
            return

        reply = questionBox(
            'Delete key', 'Delete IPNS key <b>{key}</b> ?'.format(key=keyName))

        if reply is True:
            self.app.task(self.delKey, keyName)

    def onAddKeyClicked(self):
        runDialog(AddKeyDialog, self.app, parent=self)

    @ipfsOp
    async def delKey(self, ipfsop, name):
        if await ipfsop.keysRemove(name):
            modelDelete(self.model, name)
        self.updateKeysList()

    @ipfsOp
    async def listKeys(self, ipfsop):
        keys = await ipfsop.keys()
        for key in keys:
            found = modelSearch(self.model, search=key['Name'])
            if len(found) > 0:
                continue

            nameItem = UneditableItem(key['Name'])
            nameItem.setToolTip(key['Name'])

            resolveItem = KeyResolvedItem('')
            self.model.appendRow([
                nameItem,
                KeyItem(key['Id']),
                resolveItem
            ])

            self.app.task(self.keyResolve, key, resolveItem)

    @ipfsOp
    async def keyResolve(self, ipfsop, key, item):
        if not isinstance(item, KeyResolvedItem):
            return

        now = datetime.now()

        update = False
        if item.resolvedLast is None:
            update = True

        if isinstance(item.resolvedLast, datetime):
            delta = now - item.resolvedLast
            if delta.seconds > self.resolveTimeout:
                update = True

        if update is True:
            resolved = await ipfsop.nameResolve(key['Id'])

            if isinstance(resolved, dict):
                rPath = resolved.get('Path')
                if not rPath:
                    item.setBackground(QBrush(QColor('red')))
                elif item.resolvesTo and rPath != item.resolvesTo:
                    color = QColor('#c1f0c1')
                    item.setBackground(QBrush(color))
                else:
                    item.setBackground(QBrush(Qt.NoBrush))

                if rPath and IPFSPath(rPath).valid:
                    item.resolvesTo = rPath
                    item.setText(rPath)
                    item.setToolTip(
                        "{path}\n\nResolved date: {date}".format(
                            path=rPath,
                            date=now.isoformat(sep=' ', timespec='seconds')
                        ))
            else:
                item.setText(iUnknown())

            item.resolvedLast = now

        # Ensure another one
        self.app.loop.call_later(
            self.resolveTimeout, self.app.task, self.keyResolve, key, item)

    def updateKeysList(self):
        self.app.task(self.listKeys)

    def onItemDoubleClicked(self, index):
        # Browse IPNS key associated with current item on double-click
        keyHash = self.model.data(self.model.index(index.row(), 1))
        self.gWindow.addBrowserTab().browseIpnsKey(keyHash)



import os.path
import aiofiles
import asyncio
import json
import functools

from PyQt5.QtWidgets import QToolBar
from PyQt5.QtCore import QCoreApplication
from PyQt5.Qt import QSizePolicy

from galacteek import ensure
from galacteek import log
from galacteek.ipfs import ipfsOp
from galacteek.ipfs.stat import StatInfo
from galacteek.ipfs import kilobytes
from galacteek.ipfs.cidhelpers import IPFSPath

from .helpers import getIcon
from .helpers import getMimeIcon
from .helpers import getIconFromIpfs
from .helpers import getIconFromImageData
from .helpers import getIconFromMimeType
from .helpers import getFavIconFromDir
from .widgets import HashmarkToolButton
from .widgets import IPFSObjectToolButton
from .widgets import URLDragAndDropProcessor

from .i18n import iUnknown


def iQuickAccess():
    return QCoreApplication.translate(
        'toolbarQa',
        '''<p><b>Quick Access</b> toolbar</p>
           <p>Drag and drop hashmarks and IPFS objects that
           you want to have easy access to</p>
        ''')


class QuickAccessToolBar(QToolBar, URLDragAndDropProcessor):
    """
    Quick Access toolbar, child of the main toolbar
    """

    def __init__(self, parent, configName='quickaccess1'):
        super(QuickAccessToolBar, self).__init__(parent=parent)

        self.app = QCoreApplication.instance()
        self.lock = asyncio.Lock(loop=self.app.loop)
        self.configFilePath = os.path.join(
            self.app.uiDataLocation, '{}.json'.format(configName))
        self.setObjectName('toolbarQa')
        self.setToolTip(iQuickAccess())
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.setMinimumWidth(400)
        self.setAcceptDrops(True)
        self.analyzer = self.app.rscAnalyzer
        self._config = {
            'objects': []
        }

        self.ipfsObjectDropped.connect(self.onIpfsDrop)

    @property
    def config(self):
        return self._config

    async def save(self):
        try:
            async with aiofiles.open(self.configFilePath, 'w+t') as fd:
                await fd.write(json.dumps(self.config, indent=4))
        except Exception:
            log.debug('Could not save quickaccess config')

    async def load(self):
        try:
            async with aiofiles.open(self.configFilePath, 'rt') as fd:
                data = await fd.read()
                config = json.loads(data)

            if 'objects' not in config or not isinstance(
                    config['objects'], list):
                raise ValueError('Invalid config')

            self._config = config
        except:
            self._config = {
                'objects': []
            }
            self.registerDefaults()
            await self.save()

    def registerDefaults(self):
        self.registerFromMarkMeta({
            'title': 'Hardbin'})
        self.registerFromMarkMeta({
            'description': 'Distributed wikipedia.*english'})
        self.registerFromMarkMeta({
            'title': 'IPFessay'})
        self.registerFromMarkMeta({
            'title': 'IPLD explorer'})
        self.registerFromMarkMeta({
            'title': 'markup.rocks'})

    def onIpfsDrop(self, ipfsPath):
        if str(ipfsPath) not in self.config['objects']:
            ensure(self.processObject(ipfsPath))

    async def processObject(self, ipfsPath):
        path = str(ipfsPath)

        try:
            mark = self.app.marksLocal.find(path)
            if not mark:
                # Quick access without hashmark
                await self.registerSimple(ipfsPath)
            else:
                await self.registerHashmark(mark)
        except Exception as err:
            log.debug('Error while processing object: {0} {1}'.format(
                ipfsPath, str(err)))

    def registerFromMarkMeta(self, metadata):
        mark = self.app.marksLocal.searchSingleByMetadata(metadata)
        if not mark:
            return

        if mark.path not in self.config['objects']:
            self.config['objects'].append(mark.path)

    async def findIcon(self, ipfsop, ipfsPath, rscStat, mimeType,
                       maxIconSize=kilobytes(512)):
        icon = None
        statInfo = StatInfo(rscStat)

        if statInfo.valid and mimeType.isImage:
            if statInfo.dataSmallerThan(maxIconSize):
                await ipfsop.sleep()

                data = await ipfsop.catObject(str(ipfsPath), timeout=5)
                if data:
                    icon = getIconFromImageData(data)
            else:
                icon = getMimeIcon('image/x-generic')
        elif mimeType.isDir:
            favIcon = await getFavIconFromDir(ipfsop, ipfsPath)
            if favIcon:
                icon = favIcon
            else:
                icon = getMimeIcon('inode/directory')
        else:
            await ipfsop.sleep()
            icon = getIconFromMimeType(mimeType)

        if icon is None:
            icon = getIcon('unknown-file.png')

        return icon

    @ipfsOp
    async def registerSimple(self, ipfsop, ipfsPath,
                             maxIconSize=kilobytes(512)):
        """
        Register an object in the toolbar without any hashmark associated
        """

        path = str(ipfsPath)
        result = await self.analyzer(ipfsPath)
        if result is None:
            return

        mimeType, rscStat = result

        if mimeType:
            icon = await self.findIcon(ipfsop, ipfsPath, rscStat, mimeType)
        else:
            icon = getIcon('unknown-file.png')

        button = IPFSObjectToolButton(ipfsPath, parent=self, icon=icon)
        action = self.addWidget(button)
        button.setToolTip('{0} ({1})'.format(
            str(ipfsPath), mimeType.type if mimeType else iUnknown()))
        button.clicked.connect(
            lambda: ensure(self.app.resourceOpener.open(
                str(ipfsPath), openingFrom='qa')))
        button.deleteRequest.connect(
            functools.partial(self.onDelete, button, action, str(ipfsPath)))

        if path not in self.config['objects']:
            self.config['objects'].append(path)
            await self.save()

    @ipfsOp
    async def registerHashmark(self, ipfsop, mark, maxIconSize=512 * 1024):
        """
        Register an object in the toolbar with an associated hashmark
        """

        ipfsPath = IPFSPath(mark.path)
        if not ipfsPath.valid:
            return

        mPath = str(ipfsPath)
        result = await self.analyzer(ipfsPath)
        if result is None:
            return

        mimeType, rscStat = result

        mIcon = mark.markData.get('icon', None)
        icon = None

        if mIcon and IPFSPath(mIcon).valid:
            stat = await ipfsop.objStat(mIcon)

            statInfo = StatInfo(stat)

            # Check filesize from the stat on the object
            if statInfo.valid and statInfo.dataSmallerThan(maxIconSize):
                icon = await getIconFromIpfs(ipfsop, mIcon)

                if icon is None:
                    icon = getIcon('unknown-file.png')
                else:
                    if not await ipfsop.isPinned(mIcon):
                        log.debug('Pinning icon {0}'.format(mIcon))
                        await ipfsop.ctx.pin(mIcon)
        elif mimeType:
            icon = await self.findIcon(ipfsop, ipfsPath, rscStat, mimeType)
        else:
            icon = getIcon('unknown-file.png')

        button = HashmarkToolButton(mark, parent=self, icon=icon)
        action = self.addWidget(button)
        button.setToolTip(mark.markData['metadata'].get('title', iUnknown()))
        button.clicked.connect(
            lambda: ensure(self.app.resourceOpener.open(
                str(ipfsPath), openingFrom='qa')))
        button.deleteRequest.connect(
            functools.partial(self.onDelete, button, action, mPath))

        if mPath not in self.config['objects']:
            self.config['objects'].append(mPath)
            await self.save()

    def onDelete(self, button, action, mPath):
        try:
            self._config['objects'].remove(mPath)
            self.removeAction(action)
            del button
        except:
            pass
        else:
            ensure(self.save())

    async def init(self):
        """
        Add some apps and links to the quickaccess bar
        """

        await self.load()

        for path in self.config['objects']:
            await asyncio.sleep(0.05)

            ipfsPath = IPFSPath(path)
            if not ipfsPath.valid:
                continue

            mark = self.app.marksLocal.find(str(ipfsPath))
            if mark:
                await self.registerHashmark(mark)
            else:
                await self.registerSimple(ipfsPath)

        await self.save()



from PyQt5.QtWidgets import QApplication
from PyQt5.QtWidgets import QDialogButtonBox
from PyQt5.QtWidgets import QDialog
from PyQt5.QtWidgets import QMenu
from PyQt5.QtWidgets import QAction
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QHBoxLayout
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtWidgets import QLineEdit

from PyQt5.QtCore import QRegExp
from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import QCoreApplication
from PyQt5.QtCore import Qt

from PyQt5.QtGui import QRegExpValidator

from galacteek import ensure
from galacteek.ipfs import ipfsOp

from .widgets import MarkdownInputWidget
from .helpers import runDialog
from .helpers import getMimeIcon
from .helpers import getIcon
from .helpers import messageBox


def iNewBlogPost():
    return QCoreApplication.translate('GalacteekWindow', 'New blog post')


class UserWebsiteManager(QObject):
    postError = pyqtSignal(str)

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.newBlogPostAction = QAction(
            getMimeIcon('text/html'),
            iNewBlogPost(), self,
            triggered=self.onNewPost
        )

        self.blogMenu = QMenu('Blog')
        self.blogMenu.setIcon(getIcon('blog.png'))
        self.blogMenu.addAction(self.newBlogPostAction)

        self.postError.connect(lambda err: messageBox(err))

    def onNewPost(self):
        runDialog(WebsiteAddPostDialog, manager=self)


class WebsiteAddPostDialog(QDialog):
    def __init__(self, manager=None, parent=None):
        super().__init__(parent)

        self.manager = manager

        self.setWindowTitle(iNewBlogPost())
        self.app = QApplication.instance()
        self.contents = None

        buttonBox = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel, self)

        mainLayout = QVBoxLayout()

        self.title = QLineEdit()
        regexp = QRegExp(r"[A-Za-z0-9_-\s\']+")
        self.title.setValidator(QRegExpValidator(regexp))

        titleLayout = QHBoxLayout()
        titleLayout.addWidget(QLabel('Title'))
        titleLayout.addWidget(self.title)

        self.markdownInput = MarkdownInputWidget(self)
        mainLayout.addLayout(titleLayout)
        mainLayout.addWidget(self.markdownInput)
        mainLayout.addWidget(buttonBox)
        self.setLayout(mainLayout)

        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)

        self.title.setFocus(Qt.OtherFocusReason)

    def accept(self):
        self.contents = self.markdownInput.markdownText()
        title = self.title.text()

        if len(title) == 0 or len(self.contents) == 0:
            return messageBox('Please provide a title and post body')

        self.done(1)
        ensure(self.blogPost(title, self.contents))

    @ipfsOp
    async def blogPost(self, ipfsop, title, body):
        profile = ipfsop.ctx.currentProfile
        await profile.userWebsite.blogPost(title, body)



import sys

from logbook import Logger, StreamHandler
from logbook.helpers import u
from logbook.compat import redirect_warnings
from logbook.compat import redirect_logging

mainFormatString = u(
    '[{record.time:%Y-%m-%d %H:%M:%S.%f%z}] '
    '{record.level_name}: {record.module}: {record.message}')

easyFormatString = u(
    '[{record.time: %H:%M:%S}] '
    '{record.module}: {record.message}')


def basicConfig(level='INFO', redirectLogging=False):
    handler = StreamHandler(sys.stderr, level=level, bubble=True)
    handler.format_string = mainFormatString
    handler.push_application()

    if redirectLogging:
        redirect_logging()
        redirect_warnings()


loggerMain = Logger('galacteek')
loggerUser = Logger('galacteek.user')



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'galacteek/ui/dagview.ui'
#
# Created by: PyQt5 UI code generator 5.10
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_DagViewForm(object):
    def setupUi(self, DagViewForm):
        DagViewForm.setObjectName("DagViewForm")
        DagViewForm.resize(400, 300)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(DagViewForm)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.dagPath = QtWidgets.QLabel(DagViewForm)
        self.dagPath.setText("")
        self.dagPath.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.dagPath.setObjectName("dagPath")
        self.horizontalLayout.addWidget(self.dagPath)
        self.dagControlButton = QtWidgets.QToolButton(DagViewForm)
        self.dagControlButton.setMinimumSize(QtCore.QSize(32, 32))
        self.dagControlButton.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/share/icons/ipld.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.dagControlButton.setIcon(icon)
        self.dagControlButton.setIconSize(QtCore.QSize(24, 24))
        self.dagControlButton.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        self.dagControlButton.setObjectName("dagControlButton")
        self.horizontalLayout.addWidget(self.dagControlButton)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.dagTree = QtWidgets.QTreeWidget(DagViewForm)
        self.dagTree.setColumnCount(2)
        self.dagTree.setObjectName("dagTree")
        self.dagTree.headerItem().setText(0, "1")
        self.dagTree.headerItem().setText(1, "2")
        self.dagTree.header().setVisible(False)
        self.dagTree.header().setDefaultSectionSize(200)
        self.dagTree.header().setMinimumSectionSize(180)
        self.dagTree.header().setStretchLastSection(False)
        self.verticalLayout.addWidget(self.dagTree)
        self.verticalLayout_2.addLayout(self.verticalLayout)

        self.retranslateUi(DagViewForm)

    def retranslateUi(self, DagViewForm):
        _translate = QtCore.QCoreApplication.translate
        DagViewForm.setWindowTitle(_translate("DagViewForm", "Form"))

from . import galacteek_rc



import asyncio
import functools

from PyQt5.QtWidgets import QApplication
from PyQt5.QtWidgets import QHeaderView
from PyQt5.QtWidgets import QMenu
from PyQt5.QtWidgets import QWidget
from PyQt5.QtWidgets import QDesktopWidget

from PyQt5.QtCore import Qt
from PyQt5.QtCore import QUrl
from PyQt5.QtCore import QSize
from PyQt5.QtGui import QFont

from galacteek import ensure
from galacteek.core.models.atomfeeds import AtomFeedEntryItem
from galacteek.core.models.atomfeeds import AtomFeedItem
from galacteek.dweb.page import BasePage

from . import ui_atomfeeds
from .dialogs import AddAtomFeedDialog
from .widgets import GalacteekTab
from .widgets import IPFSWebView
from .helpers import messageBox
from .helpers import runDialog


class EmptyPage(BasePage):
    def __init__(self, parent=None):
        super(EmptyPage, self).__init__('atomfeedsinit.html', parent=parent)


class AtomFeedsView(QWidget):
    def __init__(self, model, parent=None):
        super().__init__(parent)

        self.app = QApplication.instance()
        self.clipboard = self.app.appClipboard

        self.desktopWidget = QDesktopWidget()
        self.desktopGeometry = self.desktopWidget.screenGeometry()

        self.lock = asyncio.Lock()
        self.model = model
        self._offlineMode = False

        self.webView = IPFSWebView(parent=self)
        self.emptyPage = EmptyPage(self)
        self.webView.setPage(self.emptyPage)

        self.ui = ui_atomfeeds.Ui_AtomFeeds()
        self.ui.setupUi(self)

        self.ui.hLayout.addWidget(self.webView)

        self.ui.addFeedButton.clicked.connect(self.onAddFeed)

        self.ui.treeFeeds.setModel(self.model)

        self.fontFeeds = QFont('Times', 14, QFont.Bold)
        self.fontBold = QFont('Times', 12, QFont.Bold)
        self.fontRead = QFont('Times', 12)

        self.ui.treeFeeds.header().setSectionResizeMode(
            QHeaderView.ResizeToContents)
        self.ui.treeFeeds.setSortingEnabled(True)
        self.ui.treeFeeds.sortByColumn(1, Qt.DescendingOrder)

        self.ui.treeFeeds.clicked.connect(self.onItemClicked)

        self.ui.treeFeeds.setContextMenuPolicy(Qt.CustomContextMenu)
        self.ui.treeFeeds.customContextMenuRequested.connect(
            self.onContextMenu)

        self.ui.treeFeeds.setMinimumSize(QSize(
            self.desktopGeometry.width() / 3,
            self.desktopGeometry.height())
        )
        self.ui.treeFeeds.setMaximumSize(QSize(
            self.desktopGeometry.width() / 2,
            self.desktopGeometry.height())
        )
        self.webView.setMinimumSize(QSize(
            self.desktopGeometry.width() / 2,
            self.desktopGeometry.height())
        )

        self.model.feedEntryAdded.connect(self.onEntryAdded)
        self.ui.treeFeeds.setRootIndex(self.model.itemRootIdx)

    def onContextMenu(self, point):
        idx = self.ui.treeFeeds.indexAt(point)
        if not idx.isValid():
            return

        item = self.model.itemFromIndex(idx)

        if isinstance(item, AtomFeedItem):
            menu = QMenu(self)
            menu.addAction('Remove feed',
                           functools.partial(self.removeFeed, item))

            menu.exec(self.ui.treeFeeds.mapToGlobal(point))

    def removeFeed(self, feedItem):
        ensure(self.app.sqliteDb.feeds.unfollow(
            feedItem.feedId))

    def onAddFeed(self):
        def urlAccepted(dlg):
            url = dlg.textValue()
            qUrl = QUrl(url)

            if not qUrl.isValid() or not qUrl.scheme():
                return messageBox('Invalid URL')

            ensure(self.app.mainWindow.atomButton.atomFeedSubscribe(url))

        runDialog(AddAtomFeedDialog, accepted=urlAccepted)

    def onEntryAdded(self, entryItem):
        if entryItem.entry.status == entryItem.entry.ENTRY_STATUS_NEW:
            entryItem.setFont(self.fontBold)
        else:
            entryItem.setFont(self.fontRead)

        parent = entryItem.parent()
        if isinstance(parent, AtomFeedItem):
            parent.updateTitle()
            parent.setFont(self.fontFeeds)

    def onItemClicked(self, idx):
        item = self.model.itemFromIndex(idx)

        if isinstance(item, AtomFeedEntryItem):
            url = QUrl(item.entry.id)
            if url.isValid():
                self.webView.load(url)

            self.model.markEntryAsRead(item)
            item.setFont(self.fontRead)
            item.parent().updateTitle()


class AtomFeedsViewTab(GalacteekTab):
    def __init__(self, gWindow, view=None):
        super(AtomFeedsViewTab, self).__init__(gWindow)

        self.view = view if view else AtomFeedsView(
            self.app.modelAtomFeeds, parent=self)
        self.addToLayout(self.view)



import asyncio
import json
import os.path
import time

import aioipfs

from galacteek import log
from galacteek import ensure
from galacteek.ipfs.wrappers import ipfsOp


class Cancelled(Exception):
    pass


class PinningMaster(object):
    """ Pins objects on request through an async queue """

    def __init__(self, ctx, checkPinned=False, statusFilePath=None):
        self.lock = asyncio.Lock()
        self._ordersQueue = asyncio.Queue()
        self._ctx = ctx
        self._pinStatus = {}
        self._processTask = None
        self._statusFilePath = statusFilePath
        self._checkPinned = checkPinned
        self._sCleanupLast = None
        self._maxStalledMessages = 48

        ctx.app.marksLocal.markAdded.connect(self.onMarkAdded)

    @property
    def ordersQueue(self):
        return self._ordersQueue

    @property
    def ipfsCtx(self):
        return self._ctx

    @property
    def pinStatus(self):
        return self._pinStatus

    @property
    def queuesNames(self):
        return list(self.pinStatus.keys())

    def onMarkAdded(self, mPath, mData):
        if 'pin' in mData and isinstance(mData['pin'], dict):
            if mData['pin']['single'] is True:
                ensure(self.queue(mPath, False, None, qname='hashmarks'))
            elif mData['pin']['recursive'] is True:
                ensure(self.queue(mPath, True, None, qname='hashmarks'))

    def debug(self, msg, **kwargs):
        log.debug('Pinning service: {}'.format(msg), **kwargs)

    def queueStatus(self, qname):
        return self.pinStatus.get(qname, None)

    def status(self, all=False):
        export = {}

        for qname, queue in self.pinStatus.items():
            for path, pinData in queue.items():
                if pinData['pinned'] is True and all is False:
                    continue
                exportQ = export.setdefault(qname, [])
                progress = None
                if pinData['status']:
                    progress = pinData['status'].get('Progress', 'Unknown')
                exportQ.append({
                    'path': path,
                    'recursive': pinData['recursive'],
                    'pinned': pinData['pinned'],
                    'ts_queued': pinData['ts_queued'],
                    'progress': progress
                })

        return export

    def activeItemsCount(self):
        count = 0
        for qname, items in self.status().items():
            count += len(items)
        return count

    async def cleanupStatus(self, pinnedExpires=60 * 2):
        now = int(time.time())

        if isinstance(self._sCleanupLast, int):
            if now - self._sCleanupLast < 60:
                return

        toDelete = []
        try:
            with await self.lock:
                for qname, queue in self.pinStatus.items():
                    for path, pinData in queue.items():
                        if pinData['pinned'] is False:
                            continue

                        if pinData['ts_pinned'] > 0 and \
                                now - pinData['ts_pinned'] > pinnedExpires:
                            toDelete.append(path)
        except BaseException:
            self.debug('Cleanup status error')
        else:
            self._sCleanupLast = now
            for path in toDelete:
                await self.pathDelete(path)

    def _emitItemsCount(self):
        self.ipfsCtx.pinItemsCount.emit(self.activeItemsCount())

    def pathRegister(self, qname, path, recursive):
        if qname not in self.pinStatus:
            self._pinStatus[qname] = {}

        self._pinStatus[qname][path] = {
            'recursive': recursive,
            'status': None,
            'ts_queued': int(time.time()),
            'ts_pinned': 0,
            'pinned': False,
            'cancel': False
        }

        self.ipfsCtx.pinNewItem.emit(path)
        self._emitItemsCount()
        return self.pinStatus[qname][path]

    def statusFromPath(self, path, qname='default'):
        if qname in self.pinStatus:
            return self.pinStatus[qname].get(path, None)

    async def pathDelete(self, path):
        with await self.lock:
            for qname, queue in self.pinStatus.items():
                if path in queue:
                    self.debug('Deleting item {item}'.format(
                        item=self.pinStatus[qname][path]))
                    del self._pinStatus[qname][path]
                    self.ipfsCtx.pinItemRemoved.emit(qname, path)
        self._emitItemsCount()

    @ipfsOp
    async def pin(self, op, path, recursive=False, qname='default'):
        """
        Pin the object referenced by ``path``

        Returns a (path, statuscode, errmsg) tuple
        """
        self.debug('Pinning object {path} to {qname} (recursive {rec})'.format(
            path=path, rec=recursive, qname=qname))

        if self._checkPinned:
            if await op.isPinned(path):
                # Already pinned
                self.debug('Already pinned: {path}'.format(path=path))
                await self.pathDelete(path)
                return (path, 0, 'Already pinned')

        pItem = self.pathRegister(qname, path, recursive)

        try:
            stalledCn = 0
            lastPTime = None

            async for pinned in op.client.pin.add(path, recursive=recursive):
                await asyncio.sleep(0)

                now = time.time()

                if pItem['cancel'] is True:
                    raise Cancelled()

                pins = pinned.get('Pins', None)
                progress = pinned.get('Progress', None)

                if progress:
                    lastPTime = now
                    self.debug('Progress {0}: {1}'.format(path, progress))

                if pinned != pItem['status']:
                    pItem['status'] = pinned
                    self.ipfsCtx.pinItemStatusChanged.emit(qname, path, pItem)
                else:
                    if pins is None and progress is None and not lastPTime:
                        # Never received any progress status yet
                        stalledCn += 1

                    if stalledCn >= self._maxStalledMessages:
                        self.debug('{0}: stalled (removing)'.format(path))
                        await self.pathDelete(path)
                        return (path, 2, 'Stalled')

                    await asyncio.sleep(1)
        except aioipfs.APIError as err:
            self.debug('Pinning error {path}: {msg}'.format(
                path=path, msg=err.message))
            await self.pathDelete(path)
            return (path, 1, err.message)
        except asyncio.CancelledError:
            self.debug('Pinning was cancelled for {path}'.format(path=path))
            await self.pathDelete(path)
            return (path, 2, 'Cancelled')
        except Cancelled:
            await self.pathDelete(path)
            return (path, 2, 'Cancelled')
        else:
            self.debug('Queue {qname}: pinning success for {path}'.format(
                qname=qname, path=path))

            pins = pItem['status'].get('Pins', None)
            if pins and isinstance(pins, list):
                # 'Pins' is a list of CIDs, mark it as pinned
                now = int(time.time())
                pItem['pinned'] = True
                pItem['ts_pinned'] = now
                self.ipfsCtx.pinFinished.emit(path)
                self._emitItemsCount()

            return (path, 0, 'OK')

    async def queue(self, path, recursive, onSuccess, qname='default'):
        """ Queue an item for processing """
        await self.ordersQueue.put((qname, path, recursive, onSuccess))
        self.ipfsCtx.pinQueueSizeChanged.emit(self.ordersQueue.qsize())

    async def start(self):
        self._processTask = self.ipfsCtx.loop.create_task(self.process())

    async def stop(self):
        if self._processTask:
            self._processTask.cancel()
            await self._processTask
        self.saveStatus()

    def restoreStatus(self, data):
        if not isinstance(data, dict):
            return

        for qname, items in data.items():
            for item in items:
                path = item.get('path', None)
                recursive = item.get('recursive', False)

                if path:
                    self.debug('Restoring item: {path} {recursive}'.format(
                        path=path, recursive=recursive))
                    ensure(self.queue(path, recursive, None, qname=qname))

    def saveStatus(self):
        with open(self._statusFilePath, 'w+t') as fd:
            fd.write(json.dumps(self.status(), indent=4))

    def cancel(self, qname, path):
        status = self.statusFromPath(path, qname=qname)
        if status:
            status['cancel'] = True

    async def process(self):
        if os.path.exists(self._statusFilePath):
            try:
                data = None
                with open(self._statusFilePath, 'rt') as fd:
                    data = json.load(fd)
            except Exception:
                self.debug('Could not load pin status file')
            else:
                if isinstance(data, dict):
                    self.restoreStatus(data)

        try:
            while True:
                await self.cleanupStatus()

                item = await self.ordersQueue.get()
                if item is None:
                    self.debug('null item in queue')
                    continue

                self.ipfsCtx.pinQueueSizeChanged.emit(self.ordersQueue.qsize())

                try:
                    qname, path, recursive, callback = item
                    f = asyncio.ensure_future(
                        self.pin(path, recursive=recursive, qname=qname))
                    if callback:
                        f.add_done_callback(callback)
                except Exception:
                    self.debug('Invalid item in queue')
                    continue
        except asyncio.CancelledError:
            self.debug('Task was cancelled')



# Galacteek, GUI startup entry point

import asyncio
import argparse
import shutil
import subprocess
import sys
import re
from distutils.version import StrictVersion

from PyQt5.QtCore import QProcess

from galacteek import log
from galacteek import ensure
from galacteek import __version__
from galacteek.core import glogger
from galacteek.core.schemes import initializeSchemes

from galacteek.ipfs import distipfsfetch
from galacteek.ui.helpers import *  # noqa
from galacteek.ui.i18n import (
    iGoIpfsFetchTimeout,
    iGoIpfsNotFound,
    iGoIpfsTooOld,
    iGoIpfsFetchAsk,
    iGoIpfsFetchError,
    iGoIpfsFetchSuccess,
    iFsRepoMigrateNotFound
)
from galacteek import application
from galacteek.appsettings import *  # noqa

try:
    import aiomonitor
except ImportError:
    haveAiomonitor = False
else:
    haveAiomonitor = True


appStarter = None


def whichIpfs():
    return shutil.which('ipfs')


def ipfsVersion():
    try:
        p = subprocess.Popen(['ipfs', 'version', '-n'], stdout=subprocess.PIPE)
        out, err = p.communicate()
        version = out.decode().strip()
        version = re.sub('(-.*$)', '', version)
        return StrictVersion(version)
    except BaseException:
        return None


async def fetchGoIpfsWrapper(app, timeout=60 * 10):
    try:
        await asyncio.wait_for(fetchGoIpfsDist(app), timeout)
    except asyncio.TimeoutError:
        app.mainWindow.statusMessage(iGoIpfsFetchTimeout())
        return None
    except Exception:
        app.mainWindow.statusMessage(iGoIpfsFetchError())
        return None
    else:
        app.mainWindow.statusMessage(iGoIpfsFetchSuccess())
        return whichIpfs()


async def fetchGoIpfsDist(app):
    async for msg in distipfsfetch.distIpfsExtract(
            dstdir=app.ipfsBinLocation, software='go-ipfs',
            executable='ipfs', version='0.4.22', loop=app.loop,
            sslverify=app.sslverify):
        try:
            code, text = msg
            app.mainWindow.statusMessage(text)
        except Exception as e:
            app.debug(str(e))


class ApplicationStarter:
    def __init__(self, args):
        self.args = args

    def start(self):
        galacteekGui(self.args)

    def startProcess(self, args):
        p = QProcess()
        prog = self.args.binarypath if self.args.binarypath else args[0]
        p.setProgram(prog)
        if len(args) > 1:
            p.setArguments(args[1:])
        p.startDetached()


def galacteekGui(args):
    progName = args.binaryname if args.binaryname else sys.argv[0]

    if args.debug:
        glogger.basicConfig(level='DEBUG')
    else:
        glogger.basicConfig(level='INFO')

    # Initialize webengine schemes before creating the application
    initializeSchemes()

    gApp = application.GalacteekApplication(
        profile=args.profile,
        debug=args.debug,
        enableOrbital=args.enableorbital,
        sslverify=False if args.nosslverify else True,
        progName=progName,
        cmdArgs=args
    )
    loop = gApp.loop
    sManager = gApp.settingsMgr

    section = CFG_SECTION_IPFSD
    if args.apiport:
        sManager.setSetting(section, CFG_KEY_APIPORT, args.apiport)
    if args.swarmport:
        sManager.setSetting(section, CFG_KEY_SWARMPORT, args.swarmport)
    if args.gatewayport:
        sManager.setSetting(section, CFG_KEY_HTTPGWPORT, args.gatewayport)

    if sManager.isTrue(CFG_SECTION_IPFSD, CFG_KEY_ENABLED):
        fsMigratePath = shutil.which('fs-repo-migrations')
        hasFsMigrate = fsMigratePath is not None

        if hasFsMigrate is False and args.migrate is True:
            gApp.systemTrayMessage('Galacteek', iFsRepoMigrateNotFound())

        enableMigrate = hasFsMigrate is True and args.migrate is True

        # Look if we can find the ipfs executable
        ipfsPath = whichIpfs()
        minVersion = StrictVersion('0.4.7')

        if ipfsPath is None or args.forcegoipfsdl:
            if args.forcegoipfsdl:
                fetchWanted = True
            else:
                fetchWanted = questionBox('go-ipfs', iGoIpfsFetchAsk())

            def fetchFinished(fut):
                path = fut.result()
                if path is None:
                    gApp.systemTrayMessage('Galacteek',
                                           iGoIpfsFetchError())
                else:
                    gApp.startIpfsDaemon(goIpfsPath=path,
                                         migrateRepo=enableMigrate)

            if fetchWanted:
                fut = ensure(fetchGoIpfsWrapper(gApp))
                fut.add_done_callback(fetchFinished)
            else:
                gApp.systemTrayMessage('Galacteek',
                                       iGoIpfsNotFound())
        else:
            iVersion = ipfsVersion()
            if not iVersion or iVersion < minVersion:
                # warning here
                log.debug('go-ipfs version found {0} is too old'.format(
                    iVersion))
                gApp.systemTrayMessage('Galacteek',
                                       iGoIpfsTooOld())

            gApp.startIpfsDaemon(goIpfsPath=ipfsPath,
                                 migrateRepo=enableMigrate)
    else:
        ensure(gApp.updateIpfsClient())

    if args.noreleasecheck is False:
        ensure(gApp.checkReleases())

    # Monitor event loop if aiomonitor is available
    # Use the context manager so loop cleanup/close is automatic

    if args.monitor is True and haveAiomonitor:
        with aiomonitor.start_monitor(loop=loop):
            with loop:
                loop.run_forever()
    else:
        with loop:
            log.debug('Inside context manager')
            loop.run_forever()


def start():
    global appStarter
    parser = argparse.ArgumentParser()

    parser.add_argument('--version', dest='version',
                        action='store_true',
                        help='Show version number')
    parser.add_argument('--apiport', default=None,
                        help='IPFS API port number')
    parser.add_argument('--swarmport', default=None,
                        help='IPFS swarm port number')
    parser.add_argument('--gatewayport', default=None,
                        help='IPFS http gateway port number')
    parser.add_argument('--profile', default='main',
                        help='Application Profile')
    parser.add_argument('--binary-name', default=None,
                        help='Binary name', dest='binaryname')
    parser.add_argument('--binary-path', default=None,
                        help='Binary path', dest='binarypath')
    parser.add_argument(
        '--monitor',
        action='store_true',
        dest='monitor',
        help='Monitor application with aiomonitor')
    parser.add_argument(
        '--from-appimage',
        action='store_true',
        dest='appimage',
        help='Running from an AppImage')
    parser.add_argument(
        '--from-dmg',
        action='store_true',
        dest='dmg',
        help='Running from a DMG image')
    parser.add_argument(
        '--migrate',
        action='store_true',
        dest='migrate',
        help='Activate automatic repository migration')
    parser.add_argument(
        '--no-release-check',
        action='store_true',
        dest='noreleasecheck',
        help="Don't check for new releases on PyPI")
    parser.add_argument(
        '--no-ssl-verify',
        action='store_true',
        dest='nosslverify',
        help="Don't check for SSL certificate validity (ipfs-search)")
    parser.add_argument(
        '--no-ipfsscheme-mutex',
        action='store_true',
        dest='noipfsmutexlock',
        help="Don't use Qt mutexes in the native IPFS scheme handler")
    parser.add_argument(
        '--force-goipfs-download',
        action='store_true',
        dest='forcegoipfsdl',
        help="Force go-ipfs download")
    parser.add_argument(
        '--enable-orbital',
        action='store_true',
        dest='enableorbital',
        help="Enable orbit-db connector")

    parser.add_argument('-d', action='store_true',
                        dest='debug', help='Activate debugging')
    args = parser.parse_args()

    if args.version is True:
        print(__version__)
        sys.exit()

    appStarter = ApplicationStarter(args)
    appStarter.start()



import jinja2
import aiofiles
import jinja2.exceptions
import os.path
from datetime import datetime
from dateutil import parser as dateparser
from tempfile import TemporaryDirectory

from galacteek import log
from galacteek.core import isoformat

from galacteek.ipfs.cidhelpers import joinIpfs
from galacteek.ipfs.cidhelpers import cidValid
from galacteek.ipfs.cidhelpers import isIpfsPath
from galacteek.ipfs.cidhelpers import isIpnsPath
from galacteek.ipfs.wrappers import ipfsOpFn

from galacteek.dweb.markdown import markitdown


def tstodate(ts):
    try:
        date = datetime.fromtimestamp(ts)
    except TypeError:
        return ''
    else:
        return isoformat(date, timespec='seconds')


def ipfspathnorm(input):
    if isinstance(input, str):
        if cidValid(input):
            return joinIpfs(input)
        elif isIpfsPath(input) or isIpnsPath(input):
            return input


def markdownconv(text):
    return markitdown(text)


def datetimeclean(datestring):
    try:
        dt = dateparser.parse(datestring)
        if dt:
            return dt.strftime('%Y-%m-%d %H:%M')
    except:
        return None


def defaultJinjaEnv():
    # We use Jinja's async rendering
    env = jinja2.Environment(
        loader=jinja2.PackageLoader('galacteek', 'templates'),
        enable_async=True
    )
    env.filters['tstodate'] = tstodate
    env.filters['ipfspathnorm'] = ipfspathnorm
    env.filters['markdown'] = markdownconv
    env.filters['dtclean'] = datetimeclean
    return env


def renderWrapper(tmpl, **kw):
    try:
        data = tmpl.render(**kw)
    except Exception as err:
        log.debug('Error rendering jinja template',
                  exc_info=err)
        return None
    else:
        return data


async def renderTemplate(tmplname, loop=None, env=None, **kw):
    env = env if env else defaultJinjaEnv()
    tmpl = env.get_template(tmplname)
    if not tmpl:
        raise Exception('template not found')

    data = await tmpl.render_async(**kw)
    return data


@ipfsOpFn
async def ipfsRender(op, env, tmplname, **kw):
    _offline = kw.pop('offline', False)
    tmpl = env.get_template(tmplname)
    if not tmpl:
        raise Exception('template not found')

    try:
        data = await tmpl.render_async(**kw)
        entry = await op.addString(data, offline=_offline)
    except Exception as err:
        log.debug('Could not render web template {0}: {1}'.format(
            tmplname, str(err)))
        return None
    else:
        if entry:
            return entry['Hash']


@ipfsOpFn
async def ipfsRenderContained(op, env, tmplname, **kw):
    """
    Render a template, wrapping the result in a directory
    """

    tmpl = env.get_template(tmplname)
    if not tmpl:
        raise Exception('template not found')

    with TemporaryDirectory() as tmpdir:
        try:
            data = await tmpl.render_async(**kw)

            async with aiofiles.open(os.path.join(
                    tmpdir, 'index.html'), 'w+t') as fd:
                await fd.write(data)

            entry = await op.addPath(tmpdir)
        except Exception as err:
            log.debug('Could not render web template {0}: {1}'.format(
                tmplname, str(err)))
            return None
        else:
            if entry:
                return entry['Hash']



import asyncio
import functools

from PyQt5.QtWidgets import QTreeWidgetItem
from PyQt5.QtWidgets import QWidget
from PyQt5.QtWidgets import QAction
from PyQt5.QtWidgets import QPushButton
from PyQt5.QtWidgets import QMenu

from PyQt5.QtCore import QCoreApplication

from PyQt5.QtGui import QKeySequence

from galacteek import ensure
from galacteek import log
from galacteek.appsettings import *  # noqa
from galacteek.ipfs import cidhelpers
from galacteek.ipfs.ipfsops import *  # noqa
from galacteek.ipfs import ipfsOp

from galacteek.ipfs.pb import decodeUnixfsDagNode
from galacteek.ipfs.pb import unixfsDtNames
from galacteek.ipfs.pb import UNIXFS_DT_FILE
from galacteek.ipfs.pb import UNIXFS_DT_RAW

from galacteek.ipfs.dag import DAGPortal
from .helpers import *
from .i18n import *
from .widgets import GalacteekTab
from .widgets import PopupToolButton
from .colors import *

from . import ui_dagview


def iDagError(path):
    return QCoreApplication.translate(
        'DagViewForm', 'Error loading the DAG object: {0}').format(path)


def iDagInfo(obj):
    return QCoreApplication.translate('DagViewForm',
                                      'DAG object: <b>{0}</b>').format(obj)


def iDagItem(it):
    return QCoreApplication.translate('DagViewForm',
                                      'Item {0}').format(it)


class DAGViewer(GalacteekTab):
    def __init__(self, dagPath, *args, **kw):
        super(DAGViewer, self).__init__(*args, **kw)

        self.dagPath = cidhelpers.IPFSPath(dagPath, autoCidConv=True)
        self._buttons = []

        self.dagWidget = QWidget()
        self.addToLayout(self.dagWidget)
        self.ui = ui_dagview.Ui_DagViewForm()
        self.ui.setupUi(self.dagWidget)

        self.configureTree()

        self.reloadAction = QAction(getIcon('ipld.png'),
                                    'Reload',
                                    self,
                                    shortcut=QKeySequence('Ctrl+r'),
                                    triggered=self.onReload)

        self.viewParentAction = QAction(getIcon('ipld.png'),
                                        iParentDagView(),
                                        triggered=self.onViewParent)

        self.collapseTreeAction = QAction('Collapse tree',
                                          self,
                                          triggered=self.onCollapseTree)
        self.expandTreeAction = QAction('Expand tree',
                                        self,
                                        triggered=self.onExpandTree)
        self.ui.dagControlButton.setMenu(QMenu(self))

        self.ui.dagControlButton.menu().addAction(self.reloadAction)
        self.ui.dagControlButton.menu().addAction(self.viewParentAction)
        self.ui.dagControlButton.menu().addSeparator()
        self.ui.dagControlButton.menu().addAction(self.collapseTreeAction)
        self.ui.dagControlButton.menu().addAction(self.expandTreeAction)

        evfilter = IPFSTreeKeyFilter(self.ui.dagTree)
        evfilter.copyHashPressed.connect(self.onCopyItemHash)
        self.ui.dagTree.installEventFilter(evfilter)

        self.ui.dagTree.itemDoubleClicked.connect(self.onItemDoubleClicked)
        self.ui.dagTree.setHeaderLabels([iKey(), iValue(), ''])

        self.rootItem = self.ui.dagTree.invisibleRootItem()
        self.app.task(self.loadDag, self.dagPath, self.rootItem)

    def configureTree(self):
        # Resize section 1 to be slightly wider than the length of
        # a CIDv1 in base32

        self.ipldMetric = self.fontMetrics().size(0, 'IPLD rules')
        self.unixMetric = self.fontMetrics().size(0, 'IPLD rules')
        cidSize = self.fontMetrics().size(0, 'a' * (59 + 20))
        self.ui.dagTree.header().resizeSection(1, cidSize.width())

    def onReload(self):
        self.app.task(self.loadDag, self.dagPath, self.rootItem)

    def onViewParent(self):
        parent = self.dagPath.parent()
        self.app.task(self.loadDag, parent, self.rootItem)

    def onCollapseTree(self):
        self.ui.dagTree.collapseAll()

    def onExpandTree(self):
        self.ui.dagTree.expandAll()

    def onCopyItemHash(self):
        currentItem = self.ui.dagTree.currentItem()
        value = currentItem.text(1)
        if cidhelpers.cidValid(value):
            self.app.setClipboardText(value)

    def onItemDoubleClicked(self, item, col):
        if col == 1:
            text = item.text(col)
            path = cidhelpers.IPFSPath(text)
            if path.valid:
                ensure(self.app.resourceOpener.open(path))

    @ipfsOp
    async def loadDag(self, ipfsop, path, item=None):
        """
        Load the DAG from an IPFSPath

        :param IPFSPath path: object path to load
        """

        if item is None:
            item = self.rootItem

        if not path.valid:
            return messageBox(iInvalidCID())

        self.viewParentAction.setEnabled(not path.isRoot)
        self.reloadAction.setEnabled(path.isIpns)

        dagNode = await ipfsop.dagGet(path.objPath)
        if not dagNode:
            return messageBox(iDagError(path.objPath))

        await asyncio.sleep(0)

        self.dagPath = path
        self.ui.dagTree.clear()
        self.ui.dagPath.setText(iDagInfo(self.dagPath.objPath))

        await self.displayItem(dagNode, item)
        item.setExpanded(True)

        self.ui.dagTree.resizeColumnToContents(0)

    async def displayItem(self, data, item):
        if data is None:
            return

        await asyncio.sleep(0)

        if isinstance(data, dict):
            if 'data' in data and 'links' in data:
                node = decodeUnixfsDagNode(data['data'])

                if node is None:
                    dictItem = QTreeWidgetItem(item)
                    dictItem.setText(0, iUnknown())
                    return

                dType = node['type']
                dictItem = QTreeWidgetItem(item)
                dictItem.setText(0, iUnixFSNode())
                dictItem.setText(1,
                                 'Type: {type} ({typeh})'.format(
                                     type=dType,
                                     typeh=unixfsDtNames.get(dType, iUnknown())
                                 ))

                if dType in [UNIXFS_DT_FILE, UNIXFS_DT_RAW] and node['data']:
                    buttonSave = QPushButton(self)
                    buttonSave.setMaximumWidth(self.unixMetric.width())

                    buttonSave.setIcon(getIcon('save-file.png'))
                    buttonSave.clicked.connect(functools.partial(
                        self.copyFileData,
                        node['data']
                    ))

                    fileItem = QTreeWidgetItem(dictItem)
                    fileItem.setText(0, 'data')
                    fileItem.setText(1, 'Data size: {size} bytes'.format(
                        size=len(node['data'])))
                    self.ui.dagTree.setItemWidget(fileItem, 2, buttonSave)

                dictItem.setBackground(0, unixfsNodeColor)
                dictItem.setBackground(1, unixfsNodeColor)
                dictItem.setExpanded(True)

                data.pop('data')
                item = dictItem

            for dkey, dval in data.items():
                if dkey == '/':
                    # Merkle-link
                    linkValue = str(dval)

                    path = cidhelpers.IPFSPath(linkValue, autoCidConv=True)
                    if not path.valid:
                        continue

                    button = PopupToolButton(icon=getIcon('ipld.png'),
                                             parent=self)
                    button.setMaximumWidth(self.ipldMetric.width())
                    actionView = QAction(getIcon('ipld.png'),
                                         "View link's node", self,
                                         triggered=functools.partial(
                                         self.onFollowDAGLink,
                                         linkValue
                    ))
                    actionOpen = QAction(getIcon('open.png'),
                                         "Open object", self,
                                         triggered=lambda: ensure(
                                         self.app.resourceOpener.open(path)
                    ))

                    button.menu.addAction(actionView)
                    button.menu.addSeparator()
                    button.menu.addAction(actionOpen)
                    button.setDefaultAction(actionView)

                    dictItem = QTreeWidgetItem(item)
                    dictItem.setText(0, iMerkleLink())
                    dictItem.setText(1, path.rootCidRepr)
                    self.ui.dagTree.setItemWidget(dictItem, 2, button)

                    dictItem.setBackground(0, dagLinkColor)
                    dictItem.setBackground(1, dagLinkColor)
                    dictItem.setExpanded(True)
                else:
                    dictItem = QTreeWidgetItem(item)
                    dictItem.setText(0, dkey)
                    dictItem.setExpanded(True)

                    if isinstance(dval, str):
                        dictItem.setText(1, dval)
                        dictItem.setToolTip(1, dval)
                    elif isinstance(dval, int):
                        dictItem.setText(1, str(dval))
                        dictItem.setToolTip(1, str(dval))
                    else:
                        await self.displayItem(dval, dictItem)

        elif isinstance(data, list):
            for i in range(0, len(data)):
                listItem = QTreeWidgetItem(item)
                listItem.setText(0, iDagItem(i))
                listItem.setExpanded(True)
                await self.displayItem(data[i], listItem)
        elif isinstance(data, str):
            rawItem = QTreeWidgetItem(item)
            rawItem.setText(0, 'Leaf node')
            rawItem.setText(1, data[0:128])
            rawItem.setBackground(0, rawLeafColor)
            rawItem.setBackground(1, rawLeafColor)

    def copyFileData(self, data):
        path = saveFileSelect()

        if path:
            with open(path, 'w+b') as fd:
                fd.write(data)

    def onFollowDAGLink(self, cidString):
        if not cidString:
            return messageBox(iInvalidCID())

        view = DAGViewer(cidhelpers.joinIpfs(cidString), self.app.mainWindow)
        self.app.mainWindow.registerTab(
            view, iDagViewer(),
            current=True,
            icon=getIcon('ipld.png'),
            tooltip=cidString
        )


class DAGBuildingWidget(QWidget):
    """
    You would inherit this class if you have a widget that's gonna
    build a DAG from scratch or refine an existing DAG
    """

    def __init__(self, dagCid=None, offline=True, parent=None):
        super().__init__(parent)
        self.dagOffline = offline

        self._dagCid = dagCid
        self._dag = None

    @ipfsOp
    async def dagInit(self, ipfsop):
        log.debug('Creating new DAG')
        try:
            # Create empty DAG
            cid = await ipfsop.dagPut({}, offline=self.dagOffline, pin=False)
        except aioipfs.APIError:
            # 911
            log.debug('Cannot create DAG ?')
        else:
            if cidhelpers.cidValid(cid):
                self.dagCid = cid
                self._dag = DAGPortal(self.dagCid)

                await ipfsop.sleep()
                await self.dag.load()
                await self.dag.waitLoaded()



from datetime import datetime
from datetime import timezone
from dateutil import parser as duparser
from sys import version_info

from PyQt5.QtWidgets import QApplication


def runningApp():
    return QApplication.instance()


def utcDatetime():
    return datetime.now(timezone.utc)


def utcDatetimeIso():
    return utcDatetime().isoformat()


def parseDate(date):
    if isinstance(date, str):
        try:
            return duparser.parse(date)
        except Exception:
            pass


def isoformat(dt, sep=' ', timespec='seconds'):
    if version_info.major == 3 and version_info.minor < 6:
        return dt.isoformat(sep)
    elif version_info.major == 3 and version_info.minor >= 6:
        return dt.isoformat(sep, timespec=timespec)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'galacteek/ui/addhashmarkdialog.ui'
#
# Created by: PyQt5 UI code generator 5.10
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_AddHashmarkDialog(object):
    def setupUi(self, AddHashmarkDialog):
        AddHashmarkDialog.setObjectName("AddHashmarkDialog")
        AddHashmarkDialog.resize(1065, 418)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(AddHashmarkDialog.sizePolicy().hasHeightForWidth())
        AddHashmarkDialog.setSizePolicy(sizePolicy)
        AddHashmarkDialog.setMaximumSize(QtCore.QSize(16777215, 16777215))
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/share/icons/hashmarks.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        AddHashmarkDialog.setWindowIcon(icon)
        AddHashmarkDialog.setStyleSheet("")
        self.verticalLayout = QtWidgets.QVBoxLayout(AddHashmarkDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setLabelAlignment(QtCore.Qt.AlignCenter)
        self.formLayout.setFormAlignment(QtCore.Qt.AlignHCenter|QtCore.Qt.AlignTop)
        self.formLayout.setVerticalSpacing(10)
        self.formLayout.setObjectName("formLayout")
        self.hashmarksIconLabel = QtWidgets.QLabel(AddHashmarkDialog)
        self.hashmarksIconLabel.setText("")
        self.hashmarksIconLabel.setObjectName("hashmarksIconLabel")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.hashmarksIconLabel)
        self.resourceLabel = QtWidgets.QLabel(AddHashmarkDialog)
        self.resourceLabel.setMinimumSize(QtCore.QSize(800, 0))
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.resourceLabel.setFont(font)
        self.resourceLabel.setText("")
        self.resourceLabel.setObjectName("resourceLabel")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.resourceLabel)
        self.label = QtWidgets.QLabel(AddHashmarkDialog)
        self.label.setObjectName("label")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label)
        self.title = QtWidgets.QLineEdit(AddHashmarkDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.title.sizePolicy().hasHeightForWidth())
        self.title.setSizePolicy(sizePolicy)
        self.title.setMaximumSize(QtCore.QSize(500, 16777215))
        self.title.setMaxLength(128)
        self.title.setObjectName("title")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.title)
        self.label_7 = QtWidgets.QLabel(AddHashmarkDialog)
        self.label_7.setObjectName("label_7")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_7)
        self.description = QtWidgets.QTextEdit(AddHashmarkDialog)
        self.description.setMaximumSize(QtCore.QSize(600, 16777215))
        self.description.setAcceptRichText(False)
        self.description.setObjectName("description")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.description)
        self.label_6 = QtWidgets.QLabel(AddHashmarkDialog)
        self.label_6.setObjectName("label_6")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.label_6)
        self.comment = QtWidgets.QLineEdit(AddHashmarkDialog)
        self.comment.setMaximumSize(QtCore.QSize(600, 16777215))
        self.comment.setMaxLength(128)
        self.comment.setObjectName("comment")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.comment)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.formLayout.setItem(4, QtWidgets.QFormLayout.FieldRole, spacerItem)
        self.label_2 = QtWidgets.QLabel(AddHashmarkDialog)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(5, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.category = QtWidgets.QComboBox(AddHashmarkDialog)
        self.category.setMaximumSize(QtCore.QSize(600, 16777215))
        self.category.setObjectName("category")
        self.formLayout.setWidget(5, QtWidgets.QFormLayout.FieldRole, self.category)
        self.label_4 = QtWidgets.QLabel(AddHashmarkDialog)
        self.label_4.setObjectName("label_4")
        self.formLayout.setWidget(6, QtWidgets.QFormLayout.LabelRole, self.label_4)
        self.newCategory = QtWidgets.QLineEdit(AddHashmarkDialog)
        self.newCategory.setMaximumSize(QtCore.QSize(600, 16777215))
        self.newCategory.setMaxLength(128)
        self.newCategory.setObjectName("newCategory")
        self.formLayout.setWidget(6, QtWidgets.QFormLayout.FieldRole, self.newCategory)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.formLayout.setItem(7, QtWidgets.QFormLayout.FieldRole, spacerItem1)
        self.label_8 = QtWidgets.QLabel(AddHashmarkDialog)
        self.label_8.setObjectName("label_8")
        self.formLayout.setWidget(9, QtWidgets.QFormLayout.LabelRole, self.label_8)
        self.pinCombo = QtWidgets.QComboBox(AddHashmarkDialog)
        self.pinCombo.setMaximumSize(QtCore.QSize(250, 16777215))
        self.pinCombo.setObjectName("pinCombo")
        self.formLayout.setWidget(9, QtWidgets.QFormLayout.FieldRole, self.pinCombo)
        self.label_3 = QtWidgets.QLabel(AddHashmarkDialog)
        self.label_3.setText("")
        self.label_3.setObjectName("label_3")
        self.formLayout.setWidget(10, QtWidgets.QFormLayout.LabelRole, self.label_3)
        self.share = QtWidgets.QCheckBox(AddHashmarkDialog)
        self.share.setObjectName("share")
        self.formLayout.setWidget(10, QtWidgets.QFormLayout.FieldRole, self.share)
        self.verticalLayout.addLayout(self.formLayout)
        self.buttonBox = QtWidgets.QDialogButtonBox(AddHashmarkDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(AddHashmarkDialog)
        self.buttonBox.accepted.connect(AddHashmarkDialog.accept)
        self.buttonBox.rejected.connect(AddHashmarkDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(AddHashmarkDialog)

    def retranslateUi(self, AddHashmarkDialog):
        _translate = QtCore.QCoreApplication.translate
        AddHashmarkDialog.setWindowTitle(_translate("AddHashmarkDialog", "Add new hashmark"))
        self.label.setText(_translate("AddHashmarkDialog", "Title"))
        self.label_7.setText(_translate("AddHashmarkDialog", "Description"))
        self.label_6.setText(_translate("AddHashmarkDialog", "Comment"))
        self.label_2.setText(_translate("AddHashmarkDialog", "Category"))
        self.label_4.setText(_translate("AddHashmarkDialog", "or create new category, e.g news/world"))
        self.label_8.setText(_translate("AddHashmarkDialog", "Pinning"))
        self.share.setText(_translate("AddHashmarkDialog", "Share (public)"))

from . import galacteek_rc




from PyQt5.QtWidgets import QStyle
from PyQt5.QtWidgets import QProxyStyle


class GalacteekStyle(QProxyStyle):
    def __init__(self, baseStyle='Fusion'):
        super().__init__(baseStyle)

    def pixelMetric(self, metric, option, widget):
        if metric == QStyle.PM_ToolBarIconSize:
            return 24
        elif metric == QStyle.PM_TabBarIconSize:
            return 16
        elif metric == QStyle.PM_ToolBarItemSpacing:
            return 0
        elif metric == QStyle.PM_ToolBarItemMargin:
            return 0
        elif metric == QStyle.PM_TreeViewIndentation:
            return 25
        else:
            return super().pixelMetric(metric, option, widget)

    def styleHint(self, hint, option, widget, returnData):
        if hint == QStyle.SH_Menu_FlashTriggeredItem:
            return int(True)
        elif hint == QStyle.SH_ToolTip_WakeUpDelay:
            return 600
        elif hint == QStyle.SH_ToolTip_FallAsleepDelay:
            return 800
        else:
            return super().styleHint(hint, option, widget, returnData)



import functools
import aiofiles
import os.path
import json
import os
import uuid
from datetime import datetime

from PyQt5.QtWidgets import QWidget
from PyQt5.QtWidgets import QCheckBox
from PyQt5.QtWidgets import QScrollArea
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtWidgets import QHBoxLayout
from PyQt5.QtWidgets import QTextEdit
from PyQt5.QtWidgets import QSpacerItem
from PyQt5.QtWidgets import QSizePolicy
from PyQt5.QtWidgets import QToolButton
from PyQt5.QtWidgets import QPushButton
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QFrame
from PyQt5.QtWidgets import QApplication
from PyQt5.QtWidgets import QLineEdit

from PyQt5.QtCore import QDateTime
from PyQt5.QtCore import Qt
from PyQt5.QtCore import QRegularExpression
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import QDir
from PyQt5.QtCore import QTimer

from PyQt5.QtGui import QTextDocument
from PyQt5.QtGui import QTextFormat
from PyQt5.QtGui import QColor
from PyQt5.QtGui import QSyntaxHighlighter
from PyQt5.QtGui import QTextCharFormat
from PyQt5.QtGui import QFont

from galacteek import log
from galacteek.ipfs.ipfsops import *
from galacteek.core import isoformat
from galacteek.appsettings import *
from galacteek.ipfs import ipfsOp
from galacteek.ipfs.cidhelpers import IPFSPath
from galacteek.dweb.markdown import markitdown

from .helpers import *

from .ipfsview import IPFSHashExplorerWidgetFollow
from .widgets import PopupToolButton
from .widgets import IPFSPathClipboardButton
from .widgets import IPFSUrlLabel
from .widgets import GalacteekTab
from .widgets import CheckableToolButton
from .i18n import *


def iSave():
    return QCoreApplication.translate('textEditor',
                                      'Save')


def iEdit():
    return QCoreApplication.translate('textEditor',
                                      'Edit')


def iMarkdownPreview():
    return QCoreApplication.translate('textEditor',
                                      'Markdown preview')


class PythonSyntaxHighlighter(QSyntaxHighlighter):
    """
    An horrendous Python syntax highligher
    """

    def __init__(self, doc):
        super().__init__(doc)

        self.fmt1 = self.charFormat(color=Qt.red)
        self.fmt2 = self.charFormat(color=Qt.blue)
        self.fmt3 = self.charFormat(color=QColor('#AA1811'))
        self.fmt4 = self.charFormat(color=QColor('#6C0404'))
        self.fmt5 = self.charFormat(color=QColor('#65416C'))

        self.exprs = [
            (r'^(?<instr>from)\s', self.fmt1),
            (r'\s*(?<instr>import)\s', self.fmt1),
            (r'(?<instr>def) ', self.fmt3),
            (r'(?<instr>\s*(async|await|print|return|yield|pass|raise)) ',
                self.fmt3),
            (r'(?<instr>\s*(True|False|None))\s', self.fmt5),
            (r'^(?<instr>class)\s', self.fmt4),
            (r"(?<instr>('[a-zA-Z0-9_-]*'))", self.fmt4),
            (r"(?<instr>(\"[a-zA-Z0-9_-]*\"))", self.fmt4),
            (r"^((?<instr>(\s+)(if|elif|else|or|not|is)\s)", self.fmt4),
            (r"^((?<instr>(\s+)(try|except))\s", self.fmt4),
            (r"^(?<instr>(\s*)#.*$)", self.fmt5),
        ]

    def charFormat(self, color=Qt.red, fontWeight=QFont.Bold):
        fmt = QTextCharFormat()
        fmt.setFontWeight(fontWeight)
        fmt.setForeground(color)
        return fmt

    def highlightBlock(self, text):
        for expr in self.exprs:
            fmt = expr[1]
            rExpr = QRegularExpression(expr[0])

            if not rExpr.isValid():
                continue

            matches = rExpr.globalMatch(text)

            if not matches.isValid():
                continue

            while matches.hasNext():
                match = matches.next()
                instr = match.captured('instr')
                if instr:
                    self.setFormat(
                        match.capturedStart(), match.capturedLength(), fmt)


defaultStyleSheet = '''
    QTextEdit {
        padding: 5px;
        color: #242424;
        background-color: #F0F7F7;
        font: 11pt "Courier";
    }
'''


class Editor(QTextEdit):
    def __init__(self, parent, lineWrap=80, tabReplace=False):
        super(Editor, self).__init__(parent)

        self.setLineWrapMode(QTextEdit.FixedColumnWidth)
        self.setLineWrapColumnOrWidth(lineWrap)
        self.setStyleSheet(defaultStyleSheet)
        self.tabReplace = tabReplace

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Tab and self.tabReplace:
            self.insertPlainText(' ' * 4)
        else:
            super().keyPressEvent(event)

    def connectSignals(self):
        disconnectSig(self.cursorPositionChanged, self.highlightCurrentLine)
        self.cursorPositionChanged.connect(self.highlightCurrentLine)

    def highlightCurrentLine(self):
        extraSelections = []

        if not self.isReadOnly():
            selection = QTextEdit.ExtraSelection()
            lineColor = QColor('#e2dd8e')

            selection.format.setBackground(lineColor)
            selection.format.setProperty(QTextFormat.FullWidthSelection, True)
            selection.cursor = self.textCursor()
            selection.cursor.clearSelection()
            extraSelections.append(selection)

        self.setExtraSelections(extraSelections)


class EditHistory(QScrollArea):
    def __init__(self, widgetHeight=60, parent=None):
        super(EditHistory, self).__init__(parent)

        self.setSizePolicy(
            QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.setWidgetResizable(True)
        self.setMaximumHeight(widgetHeight)

        self.frame = QFrame(self)
        self.frame.setSizePolicy(
            QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.fLayout = QVBoxLayout()
        self.frame.setLayout(self.fLayout)
        self.setWidget(self.frame)


class Document(QTextDocument):
    modified = pyqtSignal()
    filenameChanged = pyqtSignal(str)

    def __init__(self, encoding=None, name=None, text='', parent=None):
        super().__init__(text, parent)

        self._filename = name
        self._changed = False
        self._encoding = encoding if encoding else 'utf-8'
        self.setDefaultStyleSheet(defaultStyleSheet)
        self.setModified(False)
        self.modificationChanged.connect(self.onModifChange)

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, name):
        self._filename = name
        self.filenameChanged.emit(name)

    @property
    def changed(self):
        return self._changed

    @property
    def encoding(self):
        return self._encoding

    def onModifChange(self, changed):
        if self.changed is False and changed is True:
            self.modified.emit()

        self._changed = changed


class TextEditorTab(GalacteekTab):
    def __init__(self, editing=False, parent=None):
        super(TextEditorTab, self).__init__(parent)

        self.editor = TextEditorWidget(
            editing=editing,
            parent=self)
        self.editor.documentChanged.connect(self.onDocChange)
        self.editor.documentNameChanged.connect(self.onDocnameChanged)

        self.vLayout.addWidget(self.editor)

    def onDocChange(self, doc):
        self.changeTabName(doc.filename)

    def changeTabName(self, name):
        if name is None:
            self.setTabName(iUnknown(), widget=self)

        elif len(name) > 16:
            cutoff = int(max(
                4 * (len(name) / 5),
                len(name) / 2
            ))

            self.setTabName(
                '... {0}'.format(name[cutoff:]),
                widget=self
            )
        else:
            self.setTabName(name, widget=self)

        self.setToolTip(name)

    def onDocnameChanged(self, doc, name):
        if doc is self.editor.currentDocument:
            self.changeTabName(doc.filename)

    def onClose(self):
        if self.editor.checkChanges() is True:
            return True
        return False


class TextEditorWidget(QWidget):
    """
    Simple Editor widget that stores the edited files in a unixfs node
    """

    rootMultihashChanged = pyqtSignal(str)
    documentChanged = pyqtSignal(Document)
    documentNameChanged = pyqtSignal(Document, str)
    filenameEntered = pyqtSignal(str)

    def __init__(self, offline=True, editing=False, sessionDagCid=None,
                 parent=None):
        super(TextEditorWidget, self).__init__(parent)

        self.app = QApplication.instance()
        self.setLayout(QVBoxLayout(self))
        self.textEditor = Editor(self)

        self.localDirName = self.genUid()
        self.localPath = os.path.join(
            self.app.tempDir.path(), self.localDirName)
        self.localDir = QDir(self.localPath)

        self._previewTimer = QTimer(self)
        self._previewTimer.timeout.connect(self.onPreviewRerender)
        self._editing = editing
        self._currentDocument = None
        self._unixDir = None

        self.setObjectName('textEditor')
        self.inPreview = False
        self.offline = offline

        self.editHistory = EditHistory(parent=self)
        self.editHistory.hide()
        self.copiesLayout = self.editHistory.fLayout

        self.ctrlLayout = QHBoxLayout()

        self.previewButton = CheckableToolButton(
            icon=getIcon('document-preview.png'),
            toggled=self.onPreviewToggled, parent=self
        )
        self.previewButton.setToolTip(iMarkdownPreview())

        self.fsViewButton = CheckableToolButton(
            icon=getIcon('folder-open.png'),
            toggled=self.onFsViewToggled, parent=self
        )
        self.fsViewButton.setEnabled(False)

        self.editButton = CheckableToolButton(
            toggled=self.onEditToggled, parent=self
        )

        self.saveButton = QToolButton(self)
        self.saveButton.setIcon(getIcon('save-file.png'))
        self.saveButton.clicked.connect(self.onSave)
        self.saveButton.setToolTip(iSave())
        self.saveButton.setEnabled(False)

        strokeIcon = getIcon('stroke-cube.png')
        self.fsControlButton = PopupToolButton(
            mode=QToolButton.InstantPopup,
            icon=strokeIcon
        )
        self.fsControlButton.setEnabled(False)

        self.fsControlButton.menu.addAction(
            strokeIcon, 'New file', self.onNewFile)

        self.editButton.setText(iEdit())

        self.unixDirCid = None
        self.unixDirDescrLabel = QLabel('Current session:')
        self.unixDirLabel = IPFSUrlLabel(
            None, invalidPathLabel='No session yet')
        self.unixDirClipButton = IPFSPathClipboardButton(None, parent=self)

        self.nameInputLabel = QLabel('File name')
        self.nameInputLine = QLineEdit('', self)
        self.nameInputLine.setMaximumWidth(160)
        self.nameInputValidate = QPushButton('OK')
        self.nameInputValidate.setMaximumWidth(40)
        self.nameInputValidate.clicked.connect(self.onNameInput)
        self.nameInputLine.returnPressed.connect(self.onNameInput)

        self.filenameEntered.connect(self.onFilenameEntered)

        self.ctrlLayout.addWidget(self.fsControlButton)
        self.ctrlLayout.addWidget(self.editButton)
        self.ctrlLayout.addWidget(self.previewButton)
        self.ctrlLayout.addWidget(self.fsViewButton)
        self.ctrlLayout.addWidget(self.saveButton)

        self.ctrlLayout.addWidget(self.nameInputLabel)
        self.ctrlLayout.addWidget(self.nameInputLine)
        self.ctrlLayout.addWidget(self.nameInputValidate)

        self.showNameInput(False)

        self.ctrlLayout.addItem(
            QSpacerItem(10, 10, QSizePolicy.Expanding, QSizePolicy.Minimum))
        self.ctrlLayout.addWidget(self.unixDirDescrLabel)
        self.ctrlLayout.addWidget(self.unixDirLabel)
        self.ctrlLayout.addWidget(self.unixDirClipButton)

        self.textVLayout = QVBoxLayout()
        self.textHLayout = QHBoxLayout()
        self.textVLayout.addLayout(self.textHLayout)

        self.previewWidget = QTextEdit(self)
        self.previewWidget.hide()

        self.currentDocument = self.newDocument()
        self.textHLayout.addWidget(self.textEditor)
        self.textHLayout.addWidget(self.previewWidget)

        optionsLayout = QHBoxLayout()
        self.tabReplaceButton = QCheckBox('Replace tabs', self)
        self.tabReplaceButton.stateChanged.connect(self.onTabReplaceToggled)
        self.tabReplaceButton.setEnabled(False)
        optionsLayout.addWidget(self.tabReplaceButton)
        self.textVLayout.addLayout(optionsLayout)

        self.layout().addLayout(self.ctrlLayout)
        self.layout().addWidget(self.editHistory)
        self.layout().addLayout(self.textVLayout)

        self.editing = editing

        self.filesView = None
        self.rootMultihashChanged.connect(self.onSessionCidChanged)

    @property
    def editing(self):
        return self._editing

    @editing.setter
    def editing(self, editing):
        if not isinstance(editing, bool):
            return

        self._editing = editing
        self.textEditor.setReadOnly(not self.editing)
        self.editButton.setChecked(self.editing)
        self.tabReplaceButton.setEnabled(self.editing)
        self.applyStylesheet()

        if self.editing is True and self.unixDirCid is None:
            ensure(self.sessionNew())

    @property
    def currentDocument(self):
        return self._currentDocument

    @currentDocument.setter
    def currentDocument(self, doc):
        self._currentDocument = doc
        self.textEditor.setDocument(self.currentDocument)

        if self.currentDocument.filename:
            if self.currentDocument.filename.endswith('.py'):
                self.highlighter = PythonSyntaxHighlighter(
                    self.currentDocument)
                self.tabReplaceButton.setChecked(True)
            if self.currentDocument.filename.endswith('.md'):
                self.previewButton.setChecked(True)
        else:
            self.highlighter = None

        self.documentChanged.emit(self.currentDocument)
        self.currentDocument.setModified(False)

        self.currentDocument.modified.connect(
            functools.partial(self.saveButton.setEnabled, True))
        self.currentDocument.contentsChanged.connect(self.onDocumentChanged)

        self.applyStylesheet()

    def onDocumentChanged(self):
        if self.previewWidget.isVisible():
            self._previewTimer.stop()
            self._previewTimer.start(1200)

    def onPreviewRerender(self):
        self.markdownPreviewUpdate()
        self._previewTimer.stop()

    def showNameInput(self, view=True):
        self.nameInputLabel.setVisible(view)
        self.nameInputLine.setVisible(view)
        self.nameInputValidate.setVisible(view)

        if view is False:
            self.nameInputLine.clear()

        if view is True:
            self.nameInputLine.setFocus(Qt.OtherFocusReason)

    def onTabReplaceToggled(self, state):
        self.textEditor.tabReplace = state

    def onNameInput(self):
        filename = self.nameInputLine.text()
        if len(filename) > 0:
            self.filenameEntered.emit(filename)

    def onFilenameEntered(self, filename):
        if self.currentDocument:
            self.currentDocument.filename = filename
            self.currentDocument.filenameChanged.emit(filename)

            self.documentNameChanged.emit(self.currentDocument, filename)
            self.showNameInput(False)
            ensure(self.saveDocument(self.currentDocument))

    def onDocnameChanged(self, doc, name):
        if doc == self.currentDocument:
            self.setTabName(name, widget=self)

    def applyStylesheet(self):
        if self.editing is True:
            self.textEditor.setStyleSheet('''
                QTextEdit {
                    background-color: #D2DBD6;
                    color: #242424;
                    padding: 5px;
                    font: 12pt "Courier";
                    font-weight: bold;
                }
            ''')
            self.textEditor.connectSignals()
        else:
            self.textEditor.highlightCurrentLine()
            self.textEditor.setStyleSheet(defaultStyleSheet)

    @property
    def unixDirCid(self):
        return self._unixDirCid

    @property
    def unixDir(self):
        return self._unixDir

    @unixDirCid.setter
    def unixDirCid(self, cid):
        self._unixDirCid = cid

    def display(self, ipfsPath):
        if not isinstance(ipfsPath, IPFSPath) or not ipfsPath.valid:
            return

        ensure(self.showFromPath(ipfsPath))

    def genUid(self):
        return str(uuid.uuid4())

    def newDocument(self, name=None, text='', encoding='utf-8'):
        self.textEditor.setStyleSheet(defaultStyleSheet)
        return Document(name=name, text=text, parent=self, encoding=encoding)

    def markdownPreviewUpdate(self):
        textData = self.currentDocument.toPlainText()
        pDocument = self.previewWidget.document()

        try:
            html = markitdown(textData)
            pDocument.setHtml(html)
        except Exception:
            pDocument.setPlainText(textData)

    def showPreview(self):
        textData = self.currentDocument.toPlainText()
        newDocument = self.newDocument(text=textData)

        try:
            html = markitdown(textData)
            newDocument.setHtml(html)
        except Exception:
            newDocument.setPlainText(textData)

        self.previewWidget.setDocument(newDocument)
        self.previewWidget.setReadOnly(True)
        self.previewWidget.show()

    def showText(self, textData, preview=False):
        if preview:
            text = textData
            newDocument = self.newDocument(text=textData)

            try:
                html = markitdown(text)
                newDocument.setHtml(html)
            except Exception:
                newDocument.setPlainText(text)

            self.textEditor.setReadOnly(True)
            return newDocument
        else:
            doc = self.newDocument(text=textData)
            self.textEditor.setReadOnly(not self.editing)
            return doc

    def onFsViewToggled(self, checked):
        self.toggleFsView(checked)

    def toggleFsView(self, view):
        if self.filesView:
            self.filesView.setVisible(view)

    def onSave(self):
        if self.currentDocument.filename is None:
            self.showNameInput(True)
        else:
            ensure(self.saveDocument(self.currentDocument))

    def nameInput(self, label='File name'):
        name = inputText(label=label)
        if name and len(name) in range(1, 256):
            return name

    def onNewDirectory(self):
        name = self.nameInput(label='Directory name')

        if name:
            self.localDir.mkdir(name)
            ensure(self.sync())

    def checkChanges(self):
        if self.currentDocument and self.currentDocument.changed is True:
            reply = questionBox(
                'Unsaved changes',
                'The current document was not saved, continue ?')
            return reply
        return True

    def onNewFile(self):
        if not self.checkChanges():
            return

        name = self.nameInput()

        if name:
            self.currentDocument = self.newDocument(name=name)
            self.editing = True

    def onEditToggled(self, checked):
        self.editing = checked

    def onPreviewToggled(self, checked):
        self.inPreview = checked

        if checked:
            self.showPreview()
        else:
            self.previewWidget.hide()

    def onSessionCidChanged(self, cid):
        log.debug('Session now at CID: {}'.format(cid))
        self.unixDirCid = cid
        self.unixDirPath = IPFSPath(cid)
        self.unixDirLabel.path = self.unixDirPath
        self.unixDirClipButton.path = self.unixDirPath

        if not self.filesView:
            self.filesView = IPFSHashExplorerWidgetFollow(
                cid, parent=self,
                addActions=False,
                autoOpenFiles=False,
                mimeDetectionMethod='magic')
            self.filesView.fileOpenRequest.connect(
                self.onFilesViewOpen)
            self.filesView.setMaximumWidth(self.width() / 3)
            self.textHLayout.addWidget(self.filesView)
            self.fsViewButton.setEnabled(True)
            self.filesView.hide()
        else:
            if self.filesView.rootHash != cid:
                self.filesView.parentHash = None
                self.filesView.changeMultihash(cid)
                self.filesView.updateTree()

    def onFilesViewOpen(self, ipfsPath):
        ensure(self.showFromPath(ipfsPath))

    @ipfsOp
    async def sessionNew(self, ipfsop):
        if not os.path.exists(self.localPath):
            try:
                os.mkdir(self.localPath)
            except:
                pass

        p = os.path.join(self.localPath, '.info')
        datet = datetime.now()
        async with aiofiles.open(p, 'w+t') as fd:
            await fd.write(json.dumps({
                'created': isoformat(datet)
            }))

        if self.currentDocument and self.currentDocument.filename:
            p = os.path.join(self.localPath, self.currentDocument.filename)
            async with aiofiles.open(p, 'w+t') as fd:
                await fd.write(self.currentDocument.toPlainText())

        await self.sync()
        self.fsControlButton.setEnabled(True)

    def makeDatetime(self, date=None):
        datet = date if date else datetime.now()
        return isoformat(datet)

    @ipfsOp
    async def sync(self, ipfsop):
        try:
            async with ipfsop.offlineMode() as opoff:
                entry = await opoff.addPath(self.localPath, wrap=False,
                                            hidden=True)
        except Exception as err:
            log.debug(str(err))
        else:
            if not entry:
                return

            self.rootMultihashChanged.emit(entry['Hash'])
            return entry

    @ipfsOp
    async def saveDocument(self, ipfsop, document, offline=True):
        text = document.toPlainText()

        if not document.filename:
            return

        docPath = os.path.join(self.localPath, document.filename)
        async with aiofiles.open(docPath, 'w+t') as fd:
            await fd.write(text)

        document.setModified(False)
        self.saveButton.setEnabled(False)

        entry = await self.sync()
        if entry:
            layout = QHBoxLayout()
            path = IPFSPath(entry['Hash'])

            urlLabel = IPFSUrlLabel(path)
            clipButton = IPFSPathClipboardButton(path)

            now = QDateTime.currentDateTime()
            date = QLabel()
            date.setText(now.toString('hh:mm:ss'))

            publishButton = QToolButton()
            publishButton.setText('Publish')
            publishButton.clicked.connect(functools.partial(
                self.onPublishFile, entry, publishButton))

            layout.addWidget(date)
            layout.addWidget(urlLabel)
            layout.addWidget(clipButton)
            layout.addItem(QSpacerItem(10, 10, QSizePolicy.Expanding,
                           QSizePolicy.Minimum))
            layout.addWidget(publishButton)

            self.copiesLayout.insertLayout(0, layout)
            self.editHistory.show()

    @ipfsOp
    async def publishEntry(self, ipfsop, entry):
        log.debug('Publishing {}'.format(entry))
        try:
            async for msg in ipfsop.client.dht.provide(entry['Hash']):
                log.debug('Provide result: {}'.format(msg))
        except Exception:
            log.debug('Publish error')
            return False
        else:
            return True

    def onPublishFile(self, entry, pButton):
        def finished(future):
            try:
                result = future.result()
                if result is True:
                    pButton.setEnabled(False)
                    pButton.setText('Published')
            except:
                pButton.setText('ERR')

        ensure(self.publishEntry(entry),
               futcallback=finished)

    @ipfsOp
    async def showFromPath(self, ipfsop, ipfsPath):
        if not self.checkChanges():
            return

        if not ipfsPath.valid:
            return

        data = await ipfsop.client.cat(ipfsPath.objPath)

        if not data:
            return messageBox('File could not be read')

        self.showNameInput(False)

        encoding, textData = self.decode(data)

        if ipfsPath.isRoot:
            basename = None
        else:
            basename = ipfsPath.basename

        doc = self.newDocument(name=basename, text=textData,
                               encoding=encoding)
        self.currentDocument = doc

    def decode(self, data):
        for enc in ['utf-8', 'latin1', 'ascii']:
            try:
                textData = data.decode(enc)
            except BaseException:
                continue
            else:
                return enc, textData
        return None, None



import asyncio
import io
import functools

import qrcode
from PIL.Image import new as NewImage
from PIL import Image

from galacteek import log
from galacteek.ipfs.cidhelpers import IPFSPath


try:
    from pyzbar.pyzbar import decode as zbar_decode
    from pyzbar import zbar_library
    zbar_library.load()
except ImportError:
    haveZbar = False
else:
    haveZbar = True


try:
    from qreader.decoder import QRDecoder
    from qreader.scanner import ImageScanner
except ImportError:
    haveQReader = False
else:
    haveQReader = True


class ImageReader:
    def _getImage(self, data):
        try:
            image = Image.open(io.BytesIO(data))
        except:
            log.debug('Cannot open image with Pillow')
            return None
        else:
            return image


class ZbarIPFSQrDecoder(ImageReader):
    """
    Decodes IPFS QR codes with the zbar library
    """

    def decode(self, data):
        """
        :param bytes data: Raw image data or Pillow image
        :rtype: list
        """

        if isinstance(data, bytes):
            image = self._getImage(data)
        elif isinstance(data, Image.Image):
            image = data
        else:
            raise Exception('Need bytes or PIL.Image')

        if image is None:
            return None

        try:
            objects = zbar_decode(image)

            urls = []
            for obj in objects:
                if not isinstance(obj.data, bytes):
                    continue
                try:
                    decoded = obj.data.decode()
                except BaseException:
                    continue

                if len(decoded) not in range(1, 1024):
                    continue

                path = IPFSPath(decoded, autoCidConv=True)
                if path.valid and path not in urls:
                    urls.append(path)

            if len(urls) > 0:  # don't return empty list
                return urls
        except Exception:
            return None


class QReaderIPFSQrDecoder(ImageReader):
    """
    Decodes IPFS QR codes with the qreader library
    """

    def decode(self, data):
        """
        Try to decode all the QR codes with IPFS addresses contained
        in an image. Returns a list of IPFS paths, or None if no URL
        was found.

        :param bytes data: Raw image data or Pillow image
        :rtype: list
        """
        # if not isinstance(data, bytes):
        #    raise Exception('Need bytes')

        if isinstance(data, bytes):
            image = self._getImage(data)
        elif isinstance(data, Image):
            image = data
        else:
            raise Exception('Need bytes or PIL.Image')

        if image is None:
            return

        try:
            imgScanner = ImageScanner(image)
            results = QRDecoder(imgScanner).get_all()

            urls = []
            for obj in results:
                if not isinstance(obj, str):
                    continue
                if len(obj) not in range(1, 1024):
                    continue

                path = IPFSPath(obj)
                if path.valid and path not in urls:
                    urls.append(path)

            if len(urls) > 0:  # don't return empty list
                return urls
        except Exception:
            return None


class IPFSQrEncoder:
    def __init__(self, maxCodes=100):
        self._codes = []
        self._maxCodes = maxCodes

    @property
    def codes(self):
        return self._codes

    def add(self, url):
        if len(self._codes) < self._maxCodes:
            self._codes.append(url)

    def _newImage(self, width, height, mode='P', fill=255, color='#000000'):
        """ Create a new Pillow image with the given params """
        img = NewImage(mode, (width, height), color=color)
        img.paste(fill, (0, 0, img.width, img.height))
        return img

    def _newQrCode(self, qrVersion=1, border=2, boxsize=10, **custom):
        """ Create a new QR code with the given params """
        return qrcode.QRCode(
            version=qrVersion,
            error_correction=qrcode.constants.ERROR_CORRECT_L,
            box_size=boxsize,
            border=border,
            **custom
        )

    def encodeUrl(self, url, fillColor='black', backColor='white',
                  qrVersion=1):
        """ Encode an URL and return a PIL image """
        qr = self._newQrCode(qrVersion=qrVersion)
        qr.add_data(url)
        qr.make()

        qrImage = qr.make_image(fill_color=fillColor, back_color=backColor)
        return qrImage.get_image()

    async def encodeAll(self, loop=None, executor=None, method='append',
                        version=12):
        """
        Runs the encoding in an asyncio executor
        """
        loop = loop if loop else asyncio.get_event_loop()
        return await loop.run_in_executor(
            executor,
            functools.partial(self._encodeAllAppend, version=version))

    def _encodeAllAppend(self, version=12):
        """
        Generate all the QR codes and embed them all into one image by
        appending. We use QR version 12 by default

        Does basically something similar to what 'convert +append' would do,
        but probably less elegantly.

        :return: an image containing all the QR codes
        :rtype: PIL.Image
        """

        baseImageWidth = 0
        baseImageHeight = 0

        # Base image sizes based on the QR count
        # The image is resized in height if needed to fit more codes

        unit = 172
        imageSizes = {
            range(1, 20): (unit * 4, unit * 4),
            range(20, 50): (unit * 6, unit * 6),
            range(50, 70): (unit * 6, unit * 8),
            range(70, 101): (unit * 8, unit * 8)
        }

        for _range, size in imageSizes.items():
            if len(self.codes) in _range:
                baseImageWidth, baseImageHeight = size
                break

        if baseImageWidth == 0 or baseImageHeight == 0:
            raise ValueError('Cannot generate image with these parameters')

        imgMosaic = self._newImage(baseImageWidth, baseImageHeight)

        lastImage = None
        posX = 0
        posY = 0
        highest = 0

        for count, url in enumerate(self.codes):
            if len(str(url)) > 1024:
                log.debug('QR#{count} ({url}): too large, ignoring'.format(
                    count=count, url=url))
                continue

            img = self.encodeUrl(url, qrVersion=version)

            if not img:
                log.debug('QR#{count} ({url}): empty image?'.format(
                    count=count, url=url))
                continue

            try:
                # Why not ?
                img = img.resize((int(img.width / 4), int(img.height / 4)))
            except:
                log.debug('QR#{count} ({url}): cannot resize image'.format(
                    count=count, url=url))
                continue

            if img.height > highest:
                highest = img.height

            if posX + img.width > imgMosaic.width:
                # Sort of a carriage return
                posY += highest
                posX = 0
                highest = 0

            if posY + img.height > imgMosaic.height:
                imgNew = self._newImage(imgMosaic.width, posY + img.height)
                imgNew.paste(imgMosaic, (0, 0))
                imgMosaic = imgNew

            log.debug('QR #{count} (size: {size}) at: {posx}:{posy}'.format(
                count=count + 1, posx=posX, posy=posY, size=len(url)))

            imgPosY = posY
            if highest > img.height:
                imgPosY = posY + int((highest - img.height) / 2)

            imgMosaic.paste(
                img, (posX, imgPosY, posX + img.width, imgPosY + img.height))

            posX += img.width
            lastImage = img

        if lastImage is not None:
            # Crop it

            imgMosaic = imgMosaic.crop(
                (0, 0, posX if posY == 0 else imgMosaic.width,
                    posY + lastImage.height)
            )

        return imgMosaic


def IPFSQrDecoder():
    # Returns a suitable QR decoder depending on the available libs
    if haveZbar:
        log.debug('Using pyzbar decoder')
        return ZbarIPFSQrDecoder()
    elif haveQReader:
        log.debug('Using qreader decoder')
        return QReaderIPFSQrDecoder()
    else:
        return None



import functools
import asyncio
import concurrent.futures
import async_timeout
import time

from web3 import Web3
from web3.providers.websocket import WebsocketProvider
from web3.auto import w3
from ens import ENS

from PyQt5.QtWidgets import QApplication
from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSignal

from galacteek import log
from galacteek import logUser
from galacteek import ensure

from .contract import ContractOperator
from galacteek.smartcontracts import getContractByName

from requests.exceptions import ConnectionError


class EthereumConnectionParams:
    def __init__(self, rpcUrl, provType='http'):
        self.provType = provType
        self.rpcUrl = rpcUrl


class EthBlock:
    def __init__(self, block):
        self._block = block


class DeployedContract:
    def __init__(self, cInterface, address):
        self.interface = cInterface
        self.address = address


class EthereumController(QObject):
    ethConnected = pyqtSignal(bool)
    ethNewBlock = pyqtSignal(str)

    def __init__(self, connParams, web3=None, loop=None,
                 executor=None, parent=None):
        super().__init__(parent)

        self.app = QApplication.instance()

        self.loop = loop if loop else asyncio.get_event_loop()
        self.executor = executor if executor else \
            concurrent.futures.ThreadPoolExecutor(max_workers=4)

        self._params = connParams
        self._web3 = web3
        self._watchTask = None
        self._blockLatest = None
        self._stop = False
        self._loadedContracts = {}

    @property
    def web3(self):
        return self._web3

    @property
    def params(self):
        return self._params

    def getWeb3(self):
        try:
            if self._params.provType == 'http':
                web3 = Web3(Web3.HTTPProvider(self.params.rpcUrl))

            elif self._params.provType == 'websocket':
                web3 = Web3(WebsocketProvider(self.params.rpcUrl))
                return web3

            web3.eth.defaultAccount = web3.eth.accounts[0]
            self._ns = ENS.fromWeb3(web3)
            return web3
        except Exception as err:
            log.debug('Cannot get Web3 connector: {}'.format(str(err)))
            return None

    async def _e(self, fn, *args, **kw):
        timeout = kw.pop('timeout', None)
        try:
            if isinstance(timeout, int):
                with async_timeout.timeout(timeout):
                    return await self.loop.run_in_executor(
                        self.executor, functools.partial(fn, *args, **kw))
            else:
                return await self.loop.run_in_executor(
                    self.executor, functools.partial(fn, *args, **kw))
        except asyncio.TimeoutError:
            log.debug('ETH Timeout')
        except ConnectionError:
            log.debug(
                'ETH Connection error while running fn: {0}'.format(fn))
            return None

    # API coroutines

    async def stop(self):
        self._stop = True

    async def start(self):
        if not self._web3:
            self._web3 = self.getWeb3()

        if self.web3 is None:
            log.debug('Web3 invalid')
            return

        if await self.connected():
            logUser.info('Ethereum: connected to {}'.format(
                self.params.rpcUrl))
            self.ethConnected.emit(True)
            # self._watchTask = await self._e(self.watchTask)
            self._watchTask = ensure(self._e(self.watchTask))
            # await self.loadContracts()
        else:
            self.ethConnected.emit(False)

    async def latestBlock(self):
        return await self._e(self.web3.eth.getBlock, 'latest')

    async def getBlock(self, block='latest'):
        return EthBlock(await self._e(self.web3.eth.getBlock, block))

    async def connected(self):
        return await self._e(self.web3.isConnected)

    async def ensAddress(self, name):
        return await self._e(self._ns.address, name)

    def watchTask(self):
        try:
            blkfilter = self.web3.eth.filter('latest')
        except Exception:
            # Node does not support filters ?
            return

        while not self._stop:
            time.sleep(5)

            for event in blkfilter.get_new_entries():
                self.eventLatestBlock(event)

    def eventLatestBlock(self, event):
        if isinstance(event, bytes):
            blockHex = w3.toHex(event)
            self.ethNewBlock.emit(blockHex)

    async def loadContractFromAddress(self, address, abi):
        def load(addr, abi):
            contract = self.web3.eth.contract(
                address=addr,
                abi=abi
            )
            return ContractOperator(self, contract)
        return await self._e(load, address, abi)

    async def loadLocalContractFromAddr(self, contractName, address):
        lContract = getContractByName(contractName)
        if not lContract:
            return

        def load(_lContract, address):
            # Blocking call
            w3Contract = _lContract.web3Contract(self.web3)
            return w3Contract(address)

        contract = await self._e(load, lContract, address)
        if contract:
            operator = lContract.module.contractOperator(self, contract,
                                                         address, parent=self)
            self._loadedContracts.setdefault(contractName, {'default': None})
            self._loadedContracts[contractName][address] = operator

            if self._loadedContracts[contractName]['default'] is None:
                self._loadedContracts[contractName]['default'] = operator

            return lContract, operator
        else:
            log.debug('Contract invalid: {}'.format(address))
            return None, None




from .objects import GObject
from .objects import pyqtSignal


from galacteek.ipfs.dag import EvolvingDAG


class DAGSignalsTower(GObject):
    dappDeployedAtCid = pyqtSignal(EvolvingDAG, str, str)


class URLSchemesTower(GObject):
    qMappingsChanged = pyqtSignal()




from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSignal

from galacteek import log


def contractDeploy(w3, contract_interface):
    try:
        if 'bytecode' in contract_interface:
            binary = contract_interface['bytecode']
        elif 'bin' in contract_interface:
            binary = contract_interface['bin']
        else:
            raise ValueError('Invalid contract interface')

        contract = w3.eth.contract(
            abi=contract_interface['abi'],
            bytecode=binary)

        txHash = contract.constructor().transact()
        address = w3.eth.getTransactionReceipt(txHash)['contractAddress']
        return address
    except Exception:
        log.debug('Failed to deploy contract')
        return None


class ContractOperator(QObject):
    contractLoaded = pyqtSignal(str)

    def __init__(self, ctrl, contract, address, parent=None):
        super(ContractOperator, self).__init__(parent)
        self.ctrl = ctrl
        self.contract = contract
        self.address = address

    async def call(self, fn, *args):
        def _callwrap(func, *args):
            result = func(*args).call()
            return result

        return await self.ctrl._e(_callwrap, fn, *args)

    async def transact(self, fn, *args):
        def _trwrap(func, *args):
            txHash = func(*args).transact()
            receipt = self.ctrl.web3.eth.waitForTransactionReceipt(txHash, 180)
            return receipt

        return await self.ctrl._e(_trwrap, fn, *args)

    async def callByName(self, fnName, *args, **kw):
        fn = getattr(self.contract.functions, fnName)
        if fn:
            return await self.call(fn, *args)

    async def callTrByName(self, fnName, *args, **kw):
        fn = getattr(self.contract.functions, fnName)
        if fn:
            return await self.transact(fn, *args)

    async def script(self, fn, *args, **kw):
        r = await self.ctrl._e(fn, self.ctrl.web3, self.contract, *args, **kw)
        return r



import json
import time
import sys
import collections
import copy
import re
import os.path
import asyncio
import uuid
from datetime import datetime

from jsonschema import validate
from jsonschema.exceptions import ValidationError

import aiofiles

from galacteek import log
from galacteek.core import utcDatetimeIso
from galacteek.core import parseDate
from galacteek.ipfs.cidhelpers import IPFSPath
from galacteek.ipfs.cidhelpers import normpPreserve

from PyQt5.QtCore import pyqtSignal, QObject


class MarksEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, IPFSHashMark):
            return obj.data
        return json.JSONEncoder.default(self, obj)


marksKey = '_marks'
pyramidsKey = '_pyramids'
pyramidsMarksKey = '_pyramidmarks'
pyramidMaxHmarksDefault = 16


def categoryValid(category):
    return re.match('^([0-9A-Za-z-_/]+)$', category) is not None


ppRe = r"^(/(ipfs|ipns)/[\w<>\:\;\,\?\!\*\%\&\=\@\$\~/\s\.\-_\\\'\(\)\+]{1,1024}$)"  # noqa

hashmarkSchema = {
    "title": "Hashmark",
    "description": "Hashmark object",
    "type": "object",
    "patternProperties": {
        ppRe: {
            "properties": {
                "metadata": {
                    "type": "object",
                    "properties": {
                        "title": {"type": "string"},
                        "description": {"type": "string"}
                    },
                    "required": ["title", "description"]
                },
                "datecreated": {"type": "string"},
                "tags": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    }
                }
            },
            "required": ["metadata", "datecreated", "tags"]
        }
    }
}


hashmarksSchema = {
    "title": "Hashmarks",
    "description": "Hashmarks collection",
    "type": "object",
    "properties": {
        "ipfsmarks": {
            "type": "object",
            "properties": {
                "categories": {
                    "$ref": "#/definitions/category"
                }
            }
        }
    },
    "definitions": {
        "category": {
            "type": "object",
            "patternProperties": {
                r"^_marks$": hashmarkSchema,
                r"^_{0}[\w0-9]{1,128}": {
                    "$ref": "#/definitions/category"
                }
            }
        }
    }
}


class IPFSHashMark(collections.UserDict):
    @property
    def markData(self):
        return self.data[self.path]

    @property
    def metadata(self):
        return self.markData['metadata']

    @property
    def title(self):
        return self.metadata['title']

    @property
    def description(self):
        return self.metadata['description']

    @property
    def comment(self):
        return self.metadata['comment']

    @property
    def datecreated(self):
        return self.markData['datecreated']

    @property
    def dtCreated(self):
        return parseDate(self.datecreated)

    @property
    def dtcreated(self):
        dtime = self.markData['datecreated']
        try:
            return datetime.strptime(dtime, '%Y-%m-%dT%H:%M:%S.%f')
        except:
            return datetime.strptime(dtime, '%Y-%m-%d %H:%M:%S')

    def addTags(self, tags):
        self.data[self.path]['tags'] += tags

    def dump(self):
        print(json.dumps(self.data, indent=4))

    def isValid(self):
        try:
            validate(self.data, hashmarkSchema)
        except ValidationError as verr:
            log.debug('Invalid JSON schema error: {}'.format(str(verr)))
            return False
        else:
            return True

    @staticmethod
    def fromJson(mPath, metadata):
        if not IPFSPath(mPath).valid or not isinstance(metadata, dict):
            return None

        hmark = IPFSHashMark({mPath: metadata})
        hmark.path = mPath
        return hmark

    @staticmethod
    def make(path, title=None, datecreated=None, share=False,
             description='', comment='', datasize=None, cumulativesize=None,
             srcplanet='Earth', tags=None,
             numlinks=None, icon=None, pinSingle=False, pinRecursive=False):
        if datecreated is None:
            datecreated = utcDatetimeIso()

        mData = IPFSHashMark({
            path: {
                'metadata': {
                    'title': title,
                    'description': description,
                    'comment': comment,
                    'srcplanet': srcplanet,
                    'datasize': datasize,
                    'cumulativesize': cumulativesize,
                    'numlinks': numlinks,
                },
                'pin': {
                    'single': pinSingle,
                    'recursive': pinRecursive,
                    'filters': []
                },
                'datecreated': datecreated,
                'tscreated': int(time.time()),
                'icon': icon,
                'share': share,
                'tags': tags if tags else [],
            }
        })

        mData.path = path
        return mData


class MultihashPyramid(collections.UserDict):
    # Types
    TYPE_STANDARD = 0

    TYPE_GALLERY = 1

    # Flags
    FLAG_MODIFIABLE_BYUSER = 0x01

    @property
    def p(self):
        return self.data[self.name]

    @property
    def marks(self):
        return self.p[pyramidsMarksKey]

    @property
    def marksCount(self):
        return len(self.marks)

    @property
    def empty(self):
        return self.marksCount == 0

    @property
    def latest(self):
        return self.p['latest']

    @latest.setter
    def latest(self, value):
        self.p['latest'] = value

    @property
    def icon(self):
        return self.p['icon']

    @property
    def type(self):
        return self.p['type']

    @property
    def uuid(self):
        return self.p['uuid']

    @property
    def description(self):
        return self.p['description']

    @property
    def ipnsKey(self):
        return self.p['ipns']['publishtokey']

    @property
    def ipnsAllowOffline(self):
        return self.p['ipns']['allowoffline']

    @property
    def ipnsLifetime(self):
        return self.p['ipns']['lifetime']

    @property
    def maxHashmarks(self):
        return self.p.get('maxhashmarks', pyramidMaxHmarksDefault)

    @staticmethod
    def make(name, description='Pyramid', icon=None, ipnskey=None,
             internal=False, publishdelay=0, comment=None, flags=0,
             allowoffline=True, lifetime='48h',
             maxhashmarks=pyramidMaxHmarksDefault,
             type=None):
        datecreated = utcDatetimeIso()
        pyramid = MultihashPyramid({
            name: {
                'ipns': {
                    'publishtokey': ipnskey,
                    'publishdelay': publishdelay,
                    'allowoffline': allowoffline,
                    'ttl': None,
                    'lifetime': lifetime
                },
                'uuid': str(uuid.uuid4()),
                'datecreated': datecreated,
                'icon': icon,
                'latest': None,
                'description': description,
                'comment': comment,
                'type': type if type else MultihashPyramid.TYPE_STANDARD,
                'internal': internal,
                'maxhashmarks': maxhashmarks,
                'flags': flags,
                pyramidsMarksKey: []  # list of hashmarks in the pyramid
            }
        })

        pyramid.name = name
        return pyramid


class QuickAccessMapping(collections.UserDict):
    """
    Mapping for the q:// scheme
    """

    @property
    def name(self):
        return self.data['name']

    @property
    def title(self):
        return self.data['title']

    @property
    def path(self):
        return self.data['mappedto']

    @property
    def ipnsFreq(self):
        return self.data['ipnsresolvefreq']

    @staticmethod
    def make(name, ipfsMPath, title=None,
             ipnsResolveFrequency=3600):
        datecreated = utcDatetimeIso()
        mapping = QuickAccessMapping({
            'name': name,
            'mappedto': str(ipfsMPath),
            'ipnsresolvefreq': ipnsResolveFrequency,
            'datecreated': datecreated,
            'hotkey': None,
            'title': title if title else name
        })
        return mapping


class IPFSMarks(QObject):
    changed = pyqtSignal()
    markDeleted = pyqtSignal(str, str)
    markAdded = pyqtSignal(str, dict)
    feedMarkAdded = pyqtSignal(str, IPFSHashMark)

    pyramidConfigured = pyqtSignal(str)
    pyramidAddedMark = pyqtSignal(str, IPFSHashMark)
    pyramidCapstoned = pyqtSignal(str)
    pyramidNeedsPublish = pyqtSignal(str, IPFSHashMark)
    pyramidChanged = pyqtSignal(str)
    pyramidEmpty = pyqtSignal(str)

    def __init__(self, path, parent=None, data=None, autosave=True,
                 backup=False):
        super().__init__(parent)

        self._path = path
        self._autosave = autosave
        self._backup = backup
        self._marks = data if data else self.load()
        self.changed.connect(self.onChanged)
        self.lastsaved = None
        self.changed.emit()

        self.pyramidCapstoned.connect(self.onPyramidCapstone)

    @property
    def path(self):
        return self._path

    @property
    def autosave(self):
        return self._autosave

    @property
    def backup(self):
        return self._backup

    @property
    def root(self):
        return self._marks

    @property
    def _rootMarks(self):
        return self.root['ipfsmarks']

    @property
    def _rootCategories(self):
        return self._rootMarks['categories']

    @property
    def _rootFeeds(self):
        return self.root['feeds']

    @property
    def _rootQMappings(self):
        return self.root['qamappings']

    def skeleton(self):
        return {
            'uuid': str(uuid.uuid4()),
            'ipfsmarks': {
                'categories': {}
            },
            'feeds': {},
            'qamappings': []
        }

    def load(self):
        if not self.path:
            return self.skeleton()
        try:
            with open(self.path, 'rt') as fd:
                marks = json.load(fd)

            if 'qamappings' not in marks:
                marks['qamappings'] = []

            if 'uuid' not in marks:
                marks['uuid'] = str(uuid.uuid4())

            return marks
        except Exception:
            marks = collections.OrderedDict()

        if 'ipfsmarks' not in marks:
            marks = self.skeleton()

        return marks

    def onChanged(self):
        if self.autosave is True:
            self.save()

    def save(self):
        """ Save synchronously """
        if not self.path:  # don't save
            return
        try:
            with open(self.path, 'w+t') as fd:
                self.serialize(fd)
        except BaseException:
            log.debug('Could not save hashmarks ({0}'.format(
                self.path))
        else:
            if not self.backup:
                return

            now = time.time()
            if not self.lastsaved or (now - self.lastsaved) > (60 * 5):
                bkpPath = '{0}.bkp'.format(self.path)

                try:
                    with open(bkpPath, 'w+t') as fd:
                        self.serialize(fd)
                except:
                    log.debug('Could not save backup file {0}'.format(bkpPath))
                else:
                    log.debug('Hashmarks backup saved: {}'.format(bkpPath))

            self.lastsaved = now

    async def saveAsync(self):
        async with aiofiles.open(self.path, 'w+t') as fd:
            await fd.write(json.dumps(self.root, indent=4, cls=MarksEncoder))
            self.lastsaved = time.time()

    def hasCategory(self, category, parent=None):
        if parent is None:
            parent = self._rootCategories
        return category in parent

    def enterCategory(self, section, create=False):
        comps = section.lstrip('/').rstrip('/').split('/')
        return self.walk(comps, create=create)

    def walk(self, path, create=True):
        """
        Walk to a category and create intermediary parents if create
        is True. path is a list of category components
        e.g ['general', 'news'] for category path /general/news
        """
        def _walk(path, parent=None):
            for p in path:
                if p.startswith('_'):
                    return
                if p in parent.keys():
                    parent = parent[p]
                elif p not in parent.keys() and create is True:
                    self.addCategory(p, parent=parent)
                    parent = parent[p]
                else:
                    return
            return parent
        return _walk(path, parent=self._rootCategories)

    def getCategories(self):
        def _list(path, parent=None):
            for p in parent.keys():
                if p.startswith('_'):
                    continue

                fullPath = path + [p]
                yield '/'.join(fullPath)
                yield from _list(fullPath, parent=parent[p])

        return sorted(list(_list([], parent=self._rootCategories)))

    def addCategory(self, category, parent=None):
        if parent is None:
            parent = self._rootCategories

        if len(category) > 256:
            return None

        if not self.hasCategory(category, parent=parent):
            parent[category] = {
                marksKey: {},
                pyramidsKey: {}
            }
            self.changed.emit()
            return parent[category]

    def getCategoryMarks(self, category, usecopy=False):
        sec = self.enterCategory(category)
        if sec:
            if usecopy:
                return copy.copy(sec[marksKey])
            else:
                return sec[marksKey]

    def isInCategory(self, category, path):
        sec = self.enterCategory(category)
        if sec:
            spath = path.rstrip('/')
            return spath in sec[marksKey] or spath + '/' in sec[marksKey]

    def getCategoryMark(self, marks, path):
        unslashed = path.rstrip('/')
        slashed = unslashed + '/'
        if unslashed in marks:
            return IPFSHashMark.fromJson(unslashed, marks[unslashed])
        elif slashed in marks:
            return IPFSHashMark.fromJson(slashed, marks[slashed])

    def getAll(self, share=False):
        cats = self.getCategories()
        _all = {}
        for cat in cats:
            catMarks = self.getCategoryMarks(cat)
            for mpath, mark in catMarks.items():
                if mark['share'] == share:
                    _all[mpath] = mark
        return _all

    def isValid(self):
        try:
            validate(self.root, hashmarksSchema)
        except ValidationError as verr:
            log.debug(
                'Hashmarks collection: invalid JSON schema error: {}'.format(
                    str(verr)))
            return False
        else:
            log.debug('Hashmarks collection: valid JSON schema !')
            return True

    async def isValidAsync(self):
        # Run JSON schema validation in an executor
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(None, self.isValid)

    async def searchAllByMetadata(self, metadata):
        # todo: deprecate searchSingleByMetadata
        if not isinstance(metadata, dict):
            raise ValueError('Metadata needs to be a dictionary')

        categories = self.getCategories()

        title = metadata.get('title')
        descr = metadata.get('description')
        path = metadata.get('path')
        comment = metadata.get('comment')

        def metaMatch(mark, field, searchQuery):
            if field not in mark['metadata'] or not isinstance(
                    mark['metadata'][field], str):
                return

            comps = searchQuery.split()
            if len(comps) == 1:
                if re.search(searchQuery, mark['metadata'][field],
                             re.IGNORECASE):
                    return True
                else:
                    return False
            else:
                lowered = mark['metadata'][field].lower()
                for comp in comps:
                    try:
                        lowered.index(comp)
                    except:
                        return False

                return True

        for cat in categories:
            await asyncio.sleep(0)

            marks = self.getCategoryMarks(cat)
            if not marks:
                continue

            for mPath, mark in marks.items():
                if not isinstance(mPath, str):
                    continue

                if 'metadata' not in mark:
                    continue

                # Remove root prefix or you'd end up with a
                # lot of stuff when searching for 'ipfs' or 'ipns'
                mPathClear = mPath.replace(
                    '/ipfs/', '').replace('/ipns/', '')

                if path and re.search(path, mPathClear):
                    yield IPFSHashMark.fromJson(mPath, mark)
                    await asyncio.sleep(0)
                    continue

                if title and metaMatch(mark, 'title', title):
                    yield IPFSHashMark.fromJson(mPath, mark)
                    await asyncio.sleep(0)
                    continue

                if descr and metaMatch(mark, 'description', descr):
                    yield IPFSHashMark.fromJson(mPath, mark)
                    await asyncio.sleep(0)
                    continue

                if comment and metaMatch(mark, 'comment', comment):
                    yield IPFSHashMark.fromJson(mPath, mark)
                    await asyncio.sleep(0)
                    continue

    def searchSingleByMetadata(self, metadata):
        if not isinstance(metadata, dict):
            raise ValueError('Metadata needs to be a dictionary')

        categories = self.getCategories()

        title = metadata.get('title', None)
        descr = metadata.get('description', None)
        path = metadata.get('path', None)

        def metaMatch(mark, field, regexp):
            try:
                if mark['metadata'][field] and re.search(
                        regexp, mark['metadata'][field]):
                    return True
            except:
                return False

        for cat in categories:
            marks = self.getCategoryMarks(cat)
            if not marks:
                continue

            for mPath, mark in marks.items():
                if 'metadata' not in mark:
                    continue

                if path and path == mPath:
                    return IPFSHashMark.fromJson(mPath, mark)

                if title and metaMatch(mark, 'title', title):
                    return IPFSHashMark.fromJson(mPath, mark)

                if descr and metaMatch(mark, 'description', descr):
                    return IPFSHashMark.fromJson(mPath, mark)

        return None

    def find(self, mpath, category=None, delete=False):
        path = normpPreserve(mpath)
        if not IPFSPath(path).valid:
            return

        categories = self.getCategories()

        for cat in categories:
            if category and cat != category:
                continue

            if self.isInCategory(cat, path):
                marks = self.getCategoryMarks(cat)
                if not marks:
                    continue

                if delete is True:
                    del marks[path]
                    self.markDeleted.emit(cat, path)
                    self.changed.emit()
                    return True

                return self.getCategoryMark(marks, path)

    def insertMark(self, mark, category):
        # Insert a mark in given category, checking of already existing mark
        # is left to the caller
        sec = self.enterCategory(category, create=True)
        if not sec:
            return False

        # Handle IPFSHashMark or tuple
        if isinstance(mark, IPFSHashMark):
            if not mark.isValid():
                return False

            if mark.path in sec[marksKey]:
                eMark = sec[marksKey][mark.path]

                if 'icon' in mark.markData:
                    eMark['icon'] = mark.markData['icon']
                return False
            sec[marksKey].update(mark)
            self.markAdded.emit(mark.path, mark.markData)
        else:
            try:
                mPath, mData = mark
                if mPath in sec[marksKey]:
                    # Patch some fields if already exists
                    eMark = sec[marksKey][mPath]
                    if 'icon' in mData:
                        eMark['icon'] = mData['icon']
                    return False
                sec[marksKey][mPath] = mData
                self.markAdded.emit(mPath, mData)
            except Exception:
                return False

        self.changed.emit()
        return True

    def add(self, mpath, title=None, category='general', share=False, tags=[],
            description=None, icon=None, pinSingle=False, pinRecursive=False):
        if not mpath:
            return False

        iPath = IPFSPath(normpPreserve(mpath))
        if not iPath.valid:
            return False

        path = str(iPath)

        sec = self.enterCategory(category, create=True)

        if not sec:
            return False

        if self.find(path):
            # We already have stored a mark with this path
            return False

        mark = IPFSHashMark.make(path,
                                 title=title,
                                 share=share,
                                 tags=tags,
                                 description=description,
                                 icon=icon,
                                 pinSingle=pinSingle,
                                 pinRecursive=pinRecursive
                                 )

        sec[marksKey].update(mark)
        self.changed.emit()
        self.markAdded.emit(path, mark.markData)

        return True

    def delete(self, path):
        return self.find(path, delete=True)

    def merge(self, oMarks, share=None, reset=False):
        count = 0
        for cat in oMarks.getCategories():
            marks = oMarks.getCategoryMarks(cat)
            for mark in marks.items():
                try:
                    mPath, mData = mark
                    if share is True and mData['share'] is False:
                        continue

                    if 'pin' in mData and reset:
                        mDataNew = copy.copy(mData)
                        mDataNew['pin']['single'] = False
                        mDataNew['pin']['recursive'] = False
                        mDataNew['share'] = False
                        self.insertMark((mPath, mDataNew), cat)
                    else:
                        self.insertMark(mark, cat)
                    count += 1
                except:
                    continue
        self.changed.emit()
        return count

    def follow(self, ipnsp, name, active=True, maxentries=4096,
               resolveevery=3600, share=False, autoPin=False):
        if ipnsp is None:
            return

        feedsSec = self._rootFeeds
        ipnsp = normpPreserve(ipnsp)

        if ipnsp in feedsSec:
            return

        feedsSec[ipnsp] = {
            'name': name,
            'active': active,
            'maxentries': maxentries,
            'resolvepolicy': 'auto',
            'resolveevery': resolveevery,
            'resolvedlast': None,
            'share': share,
            'autopin': autoPin,
            marksKey: {},
        }
        self.changed.emit()
        return feedsSec[ipnsp]

    def feedAddMark(self, ipnsp, mark):
        feeds = self._rootFeeds
        if ipnsp not in feeds:
            return False
        feed = feeds[ipnsp]
        sec = feed[marksKey]
        if mark.path in sec:
            return False

        sec.update(mark)
        self.feedMarkAdded.emit(feed['name'], mark)
        self.changed.emit()
        return True

    def getFeeds(self):
        return list(self._rootFeeds.items())

    def getFeedMarks(self, path):
        feeds = self.getFeeds()
        for fPath, fData in feeds:
            if fPath == path:
                return fData[marksKey]

    def serialize(self, fd):
        return json.dump(self.root, fd, indent=4, cls=MarksEncoder)

    def dump(self):
        print(self.serialize(sys.stdout))

    def pyramidPathFormat(self, category, name):
        return os.path.join(category, name)

    def pyramidGetLatestHashmark(self, pyramidPath):
        pyramid = self.pyramidGet(pyramidPath)
        if not pyramid:
            return None

        if pyramid.marksCount > 0:
            try:
                latest = pyramid.marks[-1]
                (mPath, mark), = latest.items()
                return IPFSHashMark.fromJson(mPath, mark)
            except:
                return None

    def onPyramidCapstone(self, pyramidPath):
        pyramid = self.pyramidGet(pyramidPath)
        latest = pyramid.latest

        if latest:
            mark = self.pyramidGetLatestHashmark(pyramidPath)
            if mark:
                self.pyramidNeedsPublish.emit(pyramidPath, mark)

    def pyramidNew(self, name, category, icon, description=None, ipnskey=None,
                   lifetime='48h', type=MultihashPyramid.TYPE_STANDARD):
        sec = self.enterCategory(category, create=True)

        if not sec:
            return False

        if pyramidsKey not in sec:
            # Update the schema
            sec[pyramidsKey] = {}

        if name not in sec[pyramidsKey]:
            pyramid = MultihashPyramid.make(
                name, description=description,
                ipnskey=ipnskey, icon=icon,
                lifetime=lifetime,
                flags=MultihashPyramid.FLAG_MODIFIABLE_BYUSER,
                type=type
            )
            sec[pyramidsKey].update(pyramid)
            self.pyramidConfigured.emit(self.pyramidPathFormat(category, name))
            self.changed.emit()
            return sec[pyramidsKey][name]

    def pyramidAccess(self, pyramidPath):
        category = os.path.dirname(pyramidPath)
        name = os.path.basename(pyramidPath)
        sec = self.enterCategory(category, create=False)
        if sec:
            if pyramidsKey not in sec:
                # Update the schema
                sec[pyramidsKey] = {}
            return sec, category, name
        return None, None, None

    def pyramidGet(self, pyramidPath):
        sec, category, name = self.pyramidAccess(pyramidPath)
        if not sec:
            return None

        if name in sec[pyramidsKey]:
            pyramid = sec[pyramidsKey][name]
            _p = MultihashPyramid({name: pyramid})
            _p.name = name
            _p.path = pyramidPath
            return _p

    def pyramidDrop(self, pyramidPath):
        sec, category, name = self.pyramidAccess(pyramidPath)

        if not sec or pyramidsKey not in sec:
            return

        if name in sec[pyramidsKey]:
            del sec[pyramidsKey][name]
            self.changed.emit()

    def pyramidAdd(self, pyramidPath, path):
        sec, category, name = self.pyramidAccess(pyramidPath)

        if not sec:
            return False

        if name in sec[pyramidsKey]:
            pyramid = sec[pyramidsKey][name]
            count = len(pyramid[pyramidsMarksKey])

            if count >= pyramid.get('maxhashmarks', pyramidMaxHmarksDefault):
                pyramid[pyramidsMarksKey].pop(0)

            exmark = self.find(path)

            if exmark:
                mark = copy.copy(exmark)
            else:
                datenowiso = utcDatetimeIso()
                mark = IPFSHashMark.make(path,
                                         title='{0}: #{1}'.format(
                                             name, count + 1),
                                         description=pyramid['description'],
                                         datecreated=datenowiso,
                                         share=False,
                                         pinSingle=True,
                                         icon=pyramid['icon'],
                                         )
            pyramid[pyramidsMarksKey].append(mark.data)
            pyramid['latest'] = path

            self.pyramidAddedMark.emit(pyramidPath, mark)
            self.pyramidChanged.emit(pyramidPath)
            self.pyramidCapstoned.emit(pyramidPath)
            self.changed.emit()

            return True

        return False

    def pyramidPop(self, pyramidPath):
        # Pop a hashmark off the list and republish

        pyramid = self.pyramidGet(pyramidPath)
        if not pyramid:
            return False

        if pyramid.marksCount > 0:
            pyramid.marks.pop()

            # Republish if there's a hashmark available
            mark = self.pyramidGetLatestHashmark(pyramidPath)
            if mark:
                pyramid.latest = mark.path
                self.pyramidNeedsPublish.emit(pyramidPath, mark)
            else:
                pyramid.latest = None

            self.pyramidChanged.emit(pyramidPath)

            if pyramid.empty:
                self.pyramidEmpty.emit(pyramidPath)

            self.changed.emit()
            return True

    async def pyramidsInit(self):
        categories = self.getCategories()
        for cat in categories:
            await asyncio.sleep(0)

            sec = self.enterCategory(cat)
            if not sec or pyramidsKey not in sec:
                continue

            pyramids = sec[pyramidsKey]
            for name, pyramid in pyramids.items():
                await asyncio.sleep(0)
                self.pyramidConfigured.emit(self.pyramidPathFormat(cat, name))

    def qaMap(self, name, mappedTo, title=None, ipnsResolveFrequency=3600):
        for m in self.qaGetMappings():
            if m.name == name:
                return False

        mapping = QuickAccessMapping.make(
            name, mappedTo, title=title,
            ipnsResolveFrequency=ipnsResolveFrequency)
        self._rootQMappings.append(mapping.data)
        self.changed.emit()
        return True

    def qaGetMappings(self):
        return [QuickAccessMapping(m) for m in copy.copy(self._rootQMappings)]



from PyQt5.QtWidgets import QToolButton
from PyQt5.QtWidgets import QApplication
from PyQt5.QtCore import QCoreApplication
from PyQt5.QtCore import pyqtSlot
from PyQt5.QtCore import pyqtSignal

from galacteek.dweb.page import IPFSPage
from galacteek.dweb.page import DWebView, WebTab
from galacteek.dweb.page import BaseHandler

from .helpers import getIcon
from .i18n import iUnknown


def iEthConnected(rpcUrl, lBlockHash):
    return QCoreApplication.translate(
        'ethereumStatusButton',
        '''Connected to the Ethereum blockchain
           RPC Url: {0}
           Latest block's hash: {1}
           ''').format(rpcUrl, lBlockHash)


def iEthNotConnected():
    return QCoreApplication.translate(
        'ethereumStatusButton',
        'Not connected to any Ethereum node')


def iEthStatus():
    return QCoreApplication.translate(
        'ethereumStatusButton',
        'Ethereum status')


class asyncSlot:
    def __init__(self, wrapped):
        self.wrapped = wrapped
        self.name = wrapped.__name__

    @pyqtSlot()
    def __get__(self, inst, owner):
        async def wrapper(*args, **kw):
            return await self.wrapped(inst, *args, **kw)


class EthereumHandler(BaseHandler):
    newBlock = pyqtSignal(str)

    def __init__(self, controller, parent=None):
        super().__init__(parent)
        self.controller = controller


class EthereumWrapperPage(IPFSPage):
    def __init__(self, controller, parent=None):
        super().__init__('ethstatus.html', parent=parent)
        self.controller = controller
        self.handler = EthereumHandler(self.controller, parent=self)
        self.register('ether', self.handler)
        self.pageCtx['controller'] = self.controller


class EthereumStatusButton(QToolButton):
    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.setEnabled(False)
        self.app = QApplication.instance()
        self.app.ethereum.ethConnected.connect(self.onEthStatus)
        self.app.ethereum.ethNewBlock.connect(self.onNewBlock)

        self.view = DWebView(parent=None)
        self.view.show()
        self.page1 = EthereumWrapperPage(self.app.ethereum, parent=self.view)
        self.view.p = self.page1

        self.statusTab = None
        self.lBlockHash = None

        self.iconEthereum = getIcon('ethereum.png')
        self.setIcon(self.iconEthereum)

    def onToggled(self, checked):
        tabName = iEthStatus()
        if checked:
            if not self.statusTab:
                self.statusTab = WebTab(self.app.mainWindow)
            self.statusTab.attach(self.view)
            self.app.mainWindow.registerTab(self.statusTab, tabName,
                                            current=True)
        elif not checked and self.statusTab:
            tab = self.app.mainWindow.findTabWithName(tabName)
            if tab:
                self.app.mainWindow.removeTabFromWidget(tab)

            self.statusTab = None

    def onEthStatus(self, connected):
        self.setEnabled(connected)
        if connected:
            self.updateToolTip()
        else:
            self.setToolTip(iEthNotConnected())

    def updateToolTip(self):
        self.setToolTip(iEthConnected(
            self.app.ethereum.params.rpcUrl,
            self.lBlockHash if self.lBlockHash else iUnknown()))

    def onNewBlock(self, blockHash):
        self.lBlockHash = blockHash
        self.updateToolTip()
        self.page1.handler.newBlock.emit(blockHash)
